# -*- coding: utf-8 -*-
"""
Viewer widgets for PowderDiffractionDatasets
"""
import pyqtgraph as pg
from PyQt5 import QtCore, QtGui, QtWidgets

from .time_series_widget import TimeSeriesWidget, pens_and_brushes


class PowderViewer(QtWidgets.QWidget):

    baseline_parameters_signal = QtCore.pyqtSignal(dict)
    peak_dynamics_roi_signal = QtCore.pyqtSignal(float, float)  # left pos, right pos

    def __init__(self, *args, **kwargs):

        super().__init__(*args, **kwargs)

        self.powder_pattern_viewer = pg.PlotWidget(
            title="Azimuthally-averaged pattern(s)",
            labels={"left": "Intensity (counts)", "bottom": "Scattering vector (1/A)"},
        )
        self.time_series_widget = TimeSeriesWidget(parent=self)

        # Peak dynamics region-of-interest
        self.peak_dynamics_region = pg.LinearRegionItem(values=(0.2, 0.3))
        self.peak_dynamics_region.sigRegionChanged.connect(self.update_peak_dynamics)

        self._text_height = 0
        self.roi_left_text = pg.TextItem("", anchor=(1, 1))
        self.roi_right_text = pg.TextItem("", anchor=(0, 1))

        layout = QtWidgets.QVBoxLayout()
        layout.addWidget(self.powder_pattern_viewer)
        layout.addWidget(self.time_series_widget)
        self.setLayout(layout)
        self.resize(self.maximumSize())

    @QtCore.pyqtSlot()
    def update_peak_dynamics(self):
        """ Update powder peak dynamics settings on demand. """
        qmin, qmax = self.peak_dynamics_region.getRegion()

        self.peak_dynamics_roi_signal.emit(qmin, qmax)

        self.roi_left_text.setText(f"{qmin:.3f}")
        self.roi_right_text.setText(f"{qmax:.3f}")
        self.roi_left_text.setPos(qmin, self._text_height)
        self.roi_right_text.setPos(qmax, self._text_height)

    @QtCore.pyqtSlot(object, object)
    def display_powder_data(self, scattering_vector, powder_data_block):
        """
        Display the radial averages of a dataset.

        Parameters
        ----------
        scattering_vector : ndarray, shape (N,) or None
            Scattering length of the radial patterns. If None, all
            viewers are cleared.
        powder_data_block : ndarray, shape (M, N) or None
            Array for which each row is an azimuthal pattern for a specific time-delay. If None, all
            viewers are cleared.
        powder_error_block : ndarray, shape (M, N) or None
            Array for which each row is the error for the corresponding azimuthal pattern. If None, error bars
            are not displayed.
        """
        if (scattering_vector is None) or (powder_data_block is None):
            self.powder_pattern_viewer.clear()
            self.time_series_widget.clear()
            return

        pens, brushes = pens_and_brushes(num=powder_data_block.shape[0])

        self.powder_pattern_viewer.enableAutoRange()
        self.powder_pattern_viewer.clear()

        for pen, brush, curve in zip(pens, brushes, powder_data_block):
            self.powder_pattern_viewer.plot(
                scattering_vector,
                curve,
                pen=None,
                symbol="o",
                symbolPen=pen,
                symbolBrush=brush,
                symbolSize=3,
            )

        # Calculate the height of the text
        # so that it never obscures data
        self._text_height = 1.1 * powder_data_block.max()

        # clearing the powder_pattern_viewer removes all items from the view box
        self.powder_pattern_viewer.addItem(self.peak_dynamics_region)
        self.powder_pattern_viewer.addItem(self.roi_left_text)
        self.powder_pattern_viewer.addItem(self.roi_right_text)

        self.peak_dynamics_region.setBounds(
            [scattering_vector.min(), scattering_vector.max()]
        )
        self.update_peak_dynamics()  # Update peak dynamics plot if background has been changed, for example

    @QtCore.pyqtSlot(object, object)
    def display_peak_dynamics(self, times, intensities):
        """
        Display the time series associated with the integral between the bounds
        of the ROI
        """
        self.time_series_widget.plot(times, intensities)



# -*- coding: utf-8 -*-
"""
Raw dataset classes
===================
"""
from abc import abstractmethod
from collections import OrderedDict
from contextlib import AbstractContextManager
from functools import wraps, partial
from pathlib import Path
import multiprocessing as mp

import numpy as np

from npstreams import average, itercopy, peek
from skued import ialign

from .meta import ExperimentalParameter, MetaRawDataset


def open_raw(path):
    """
    Open a raw data item, guessing the AbstractRawDataset instance that
    should be used based on available plug-ins.

    This function can also be used as a context manager::

        with open_raw('.') as dset:
            ...

    Parameters
    ----------
    path : path-like
        Path to the file/folder containing the raw data.

    Returns
    -------
    raw : AbstractRawDataset instance
        The raw dataset. If no format could be guessed, an RuntimeError is raised.

    Raises
    ------
    RuntimeError
        if the data format could not be guessed.
    """
    if isinstance(path, Path):
        path = str(path)

    # For easier debugging, data formats are checked in deterministic order
    for dataformat in sorted(AbstractRawDataset.implementations, key=str):
        try:
            return dataformat(path)
        except:
            pass

    raise RuntimeError(
        f"No data format could be guessed for item located at: \n {path}"
    )


class AbstractRawDataset(AbstractContextManager, metaclass=MetaRawDataset):
    """
    Abstract base class for ultrafast electron diffraction data set.
    AbstractRawDataset allows for enforced metadata types and values,
    as well as a standard interface. For example, AbstractRawDataset
    implements the context manager interface.

    Minimally, the following method must be implemented in subclasses:

        * raw_data

    It is suggested to also implement the following magic method:

        * __init__
        * __exit__

    Optionally, the ``display_name`` class attribute can be specified.

    For better results or performance during reduction, the following methods
    can be specialized:

        * reduced

    A list of concrete implementations of AbstractRawDatasets is available in
    the ``implementations`` class attribute. Subclasses are automatically added.

    The call signature must remain the same for all overwritten methods.
    """

    # List of valid metadata below
    # Using the ExperimentalParameter allows for automatic registering
    # of the parameters as valid.
    # These attributes can be accessed using the usual property access
    date = ExperimentalParameter("date", str, default="")
    energy = ExperimentalParameter("energy", float, default=90)  # keV
    pump_wavelength = ExperimentalParameter(
        "pump_wavelength", int, default=800
    )  # nanometers
    fluence = ExperimentalParameter("fluence", float, default=0)  # mj / cm**2
    time_zero_shift = ExperimentalParameter(
        "time_zero_shift", float, default=0
    )  # picoseconds
    temperature = ExperimentalParameter("temperature", float, default=293)  # Kelvins
    exposure = ExperimentalParameter("exposure", float, default=1)  # seconds
    resolution = ExperimentalParameter("resolution", tuple, default=(2048, 2048))
    time_points = ExperimentalParameter(
        "time_points", tuple, default=tuple()
    )  # picoseconds
    scans = ExperimentalParameter("scans", tuple, default=(1,))
    camera_length = ExperimentalParameter(
        "camera_length", float, default=0.23
    )  # meters
    pixel_width = ExperimentalParameter("pixel_width", float, default=14e-6)  # meters
    notes = ExperimentalParameter("notes", str, default="")

    def __init__(self, source=None, metadata=None):
        """
        Parameters
        ----------
        source : object
            Data source, for example a directory or external file.
        metadata : dict or None, optional
            Metadata and experimental parameters. Dictionary keys that are
            not valid metadata, they are ignored. Metadata can also be
            set directly later.

        Raises
        ------
        TypeError
            if an item from the metadata has an unexpected type.
        """
        self.source = source
        if metadata:
            self.update_metadata(metadata)

    def __exit__(self, *exc):
        pass

    def __repr__(self):
        rep = f"< {type(self).__name__} object with following metadata: "
        for key, val in self.metadata.items():
            rep += f"\n    {key}: {val}"

        return rep + " >"

    def update_metadata(self, metadata):
        """
        Update metadata from a dictionary. Only appropriate keys are used; irrelevant keys are ignored.

        Parameters
        ----------
        metadata : dictionary
            See ``AbstractRawDataset.valid_metadata`` for valid keys.
        """
        for k, v in metadata.items():
            if k in self.valid_metadata:
                setattr(self, k, v)

    @property
    def metadata(self):
        """ Experimental parameters and dataset metadata as a dictionary. """
        meta = {k: getattr(self, k) for k in self.valid_metadata}
        # Ordered dictionary by keys is easiest to inspect
        return OrderedDict(sorted(meta.items(), key=lambda t: t[0]))

    def iterscan(self, scan, **kwargs):
        """
        Generator function of diffraction patterns as part of a scan, in
        time-delay order.

        Parameters
        ----------
        scan : int
            Scan from which to yield the data.
        kwargs
            Keyword-arguments are passed to ``raw_data`` method.

        Yields
        ------
        data : `~numpy.ndarray`, ndim 2

        See Also
        --------
        itertime : generator of diffraction patterns for a single time-delay, in scan order
        """
        if scan not in set(self.scans):
            raise ValueError(f"There is no scan {scan} in available scans")

        for timedelay in self.time_points:
            yield self.raw_data(timedelay=timedelay, scan=scan, **kwargs)

    def itertime(self, timedelay, exclude_scans=None, **kwargs):
        """
        Generator function of diffraction patterns of the same time-delay, in
        scan order.

        Parameters
        ----------
            timedelay : float
            Scan from which to yield the data.
        exclude_scans : iterable or None, optional
            These scans will be skipped.
        kwargs
            Keyword-arguments are passed to ``raw_data`` method.

        Yields
        ------
        data : `~numpy.ndarray`, ndim 2

        See Also
        --------
        iterscan : generator of diffraction patterns for a single scan, in time-delay order
        """
        if not exclude_scans:
            exclude_scans = set([])

        if timedelay not in set(self.time_points):
            raise ValueError(
                f"There is no time-delay {timedelay} in available time-delays"
            )

        valid_scans = sorted(set(self.scans) - set(exclude_scans))
        for scan in valid_scans:
            yield self.raw_data(timedelay=timedelay, scan=scan, **kwargs)

    @abstractmethod
    def raw_data(self, timedelay, scan=1, **kwargs):
        """
        Returns an array of the image at a timedelay and scan.

        Parameters
        ----------
        timdelay : float
            Acquisition time-delay.
        scan : int, optional
            Scan number. Default is 1.
        kwargs
            Keyword-arguments are ignored.

        Returns
        -------
        arr : `~numpy.ndarray`, ndim 2

        Raises
        ------
        ValueError
            if ``timedelay`` or ``scan`` are invalid / out of bounds.
        IOError
            Filename is not associated with an image/does not exist.
        """
        pass

    def reduced(
        self,
        exclude_scans=None,
        align=True,
        normalize=True,
        mask=None,
        processes=1,
        dtype=float,
    ):
        """
        Generator of reduced dataset. The reduced diffraction patterns are generated in order of time-delay.

        This particular implementation normalizes diffracted intensity of pictures acquired at the same time-delay
        while rejecting masked pixels.

        Parameters
        ----------
        exclude_scans : iterable or None, optional
            These scans will be skipped when reducing the dataset.
        align : bool, optional
            If True (default), raw diffraction patterns will be aligned using the masked normalized
            cross-correlation approach. See `skued.align` for more information.
        normalize : bool, optional
            If True (default), equivalent diffraction pictures (e.g. same time-delay, different scans)
            are normalized to the same diffracted intensity.
        mask : array-like of bool or None, optional
            If not None, pixels where ``mask = True`` are ignored for certain operations (e.g. alignment).
        processes : int or None, optional
            Number of Processes to spawn for processing.
        dtype : numpy.dtype or None, optional
            Reduced patterns will be cast to ``dtype``.

        Yields
        ------
        pattern : `~numpy.ndarray`, ndim 2
        """
        # Convention for masks is different for scikit-ued
        # For backwards compatibility, we cannot change the definition
        # in iris-ued
        valid_mask = np.logical_not(mask)

        kwargs = {
            "raw": self,
            "exclude_scans": exclude_scans,
            "align": align,
            "normalize": normalize,
            "valid_mask": valid_mask,
            "dtype": dtype,
        }

        combined = pmap(_raw_combine, iterable=self.time_points, kwargs=kwargs)

        # Each image at the same time-delay are aligned to each other. This means that
        # the reference image is different for each time-delay. We align the reduced images
        # to each other as well.
        if align:
            yield from ialign(combined, mask=valid_mask)
        else:
            yield from combined


# For multiprocessing, the function to be mapped must be
# global, hence defined outside of the class method
def _raw_combine(timedelay, raw, exclude_scans, normalize, align, valid_mask, dtype):

    images = raw.itertime(timedelay, exclude_scans=exclude_scans)

    if align:
        images = ialign(images, mask=valid_mask)

    # Set up normalization
    if normalize:
        images, images2 = itercopy(images, copies=2)

        # Compute the total intensity of first image
        # This will be the reference point
        first2, images2 = peek(images2)
        initial_weight = np.sum(first2[valid_mask])
        weights = (initial_weight / np.sum(image[valid_mask]) for image in images2)
    else:
        weights = None
    print("timedelay")
    return average(images, weights=weights).astype(dtype)


def check_raw_bounds(method):
    """
    Decorator that automatically checks out-of-bounds errors while
    querying raw diffraction data.

    In case of out-of-bounds queries, a ``ValueError`` is raised.

    ``method`` is expected to be a reimplemented :meth:`AbstractRawDataset.raw_data` method.
    See :meth:`AbstractRawDataset.raw_data` for the expected call signature.
    """
    # Note that we use the expected call signature from AbstractRawDataset.raw_data
    @wraps(method)
    def checked_method(self, timedelay, scan, *args, **kwargs):
        timedelay = float(timedelay)
        scan = int(scan)

        valid_scan = scan in self.scans
        valid_timedelay = timedelay in self.time_points

        if (not valid_scan) or (not valid_timedelay):
            raise ValueError(
                f"Requested time-delay {timedelay} and scan {scan} are invalid or out-of-bounds"
            )

        return method(self, timedelay, scan, *args, **kwargs)

    return checked_method


# We explicitly control the multiprocessing Pool
# rather than use npstreams.pmap because we don't
# want the process pool to precompute values like
# npstreams.pmap does via chunking
def pmap(func, iterable, args=None, kwargs=None, processes=1):
    if kwargs is None:
        kwargs = dict()

    if args is None:
        args = tuple()

    func = partial(func, *args, **kwargs)

    if processes == 1:
        yield from map(func, iterable)
        return

    with Pool(processes) as pool:
        yield from pool.imap(func=func, iterable=iterable, chunksize=1)



# -*- coding: utf-8 -*-
"""
Controller object to handle DiffractionDataset objects
"""
import traceback
import warnings
from contextlib import suppress
from functools import wraps
from shutil import copy2
from types import FunctionType
import logging

import numpy as np
from PyQt5 import QtCore
from skued import bragg_peaks
from .. import AbstractRawDataset, DiffractionDataset, PowderDiffractionDataset
from .qlogger import QLogger


def error_aware(func):
    """
    Wrap an instance method with a try/except and emit a message.
    Instance must have a signal called 'error_message_signal' which
    will be emitted with the message upon error.

    Keyboard interrupts are never ignored.
    """

    @wraps(func)
    def aware_func(self, *args, **kwargs):
        try:
            return func(self, *args, **kwargs)
        except KeyboardInterrupt:
            raise
        except Exception:
            exc = traceback.format_exc()
            self.error_message_signal.emit(exc)
            warnings.warn(exc, UserWarning)

    return aware_func


class ErrorAware(type(QtCore.QObject)):
    """ Metaclass for error-aware Qt applications. """

    def __new__(meta, classname, bases, class_dict):
        new_class_dict = dict()
        for attr_name, attr in class_dict.items():
            if isinstance(attr, FunctionType):
                attr = error_aware(attr)
            new_class_dict[attr_name] = attr

        return super().__new__(meta, classname, bases, new_class_dict)


def indicate_in_progress(method):
    """Decorator for IrisController methods that should
    emit the ``operation_in_progress`` signal and automatically
    revert when the operation is finished."""

    @wraps(method)
    def new_method(*args, **kwargs):
        args[0].operation_in_progress.emit(True)
        try:
            return method(*args, **kwargs)
        finally:
            args[0].operation_in_progress.emit(False)

    return new_method


class IrisController(QtCore.QObject, metaclass=ErrorAware):
    """
    Controller behind Iris.
    """

    raw_dataset_loaded_signal = QtCore.pyqtSignal(bool)
    processed_dataset_loaded_signal = QtCore.pyqtSignal(bool)
    powder_dataset_loaded_signal = QtCore.pyqtSignal(bool)

    raw_dataset_metadata = QtCore.pyqtSignal(dict)
    dataset_metadata = QtCore.pyqtSignal(dict)
    powder_dataset_metadata = QtCore.pyqtSignal(dict)

    error_message_signal = QtCore.pyqtSignal(str)
    processing_data_signal = QtCore.pyqtSignal(bool)
    operation_in_progress = QtCore.pyqtSignal(bool)
    status_message_signal = QtCore.pyqtSignal(str)

    raw_data_signal = QtCore.pyqtSignal(object)
    averaged_data_signal = QtCore.pyqtSignal(object, bool)
    powder_data_signal = QtCore.pyqtSignal(object, object)
    bragg_peaks_signal = QtCore.pyqtSignal(list)

    time_series_signal = QtCore.pyqtSignal(object, object)
    powder_time_series_signal = QtCore.pyqtSignal(object, object)

    relative_powder_enable_signal = QtCore.pyqtSignal(bool)
    relative_averaged_enable_signal = QtCore.pyqtSignal(bool)
    powder_bgr_enable_signal = QtCore.pyqtSignal(bool)

    processing_progress_signal = QtCore.pyqtSignal(int)
    powder_promotion_progress = QtCore.pyqtSignal(int)
    angular_average_progress = QtCore.pyqtSignal(int)

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.worker = None
        self.raw_dataset = None
        self.dataset = None

        # Internal state for powder background removal. If True, display_powder_data
        # will send background-subtracted data
        self._bgr_powder = False

        # Internal state for display format of averaged and powder data
        self._relative_powder = False
        self._relative_averaged = False
        self._timedelay_index = 0

        # array containers
        # Preallocation is important for arrays that change often. Then,
        # we can take advantage of the out parameter
        # These attributes are not initialized to None since None is a valid
        # out parameter
        # TODO: is it worth it?
        self._averaged_data_container = None
        self._average_time_series_container = None

        self.logger = QLogger(parent=self)
        self.logger.debug("Controller started.")

        # A lot of info-level messages should be mirrored between the status
        # bar and the info log
        # Similarly, the error messages from ErrorAware type should be
        # mirrored in the error log
        self.status_message_signal.connect(self.logger.info)
        self.error_message_signal.connect(self.logger.error)

    @QtCore.pyqtSlot(str)
    @QtCore.pyqtSlot(str, object)
    def log(self, message, level=logging.INFO):
        """
        Interface to the iris logger.

        ALL logging should go through the controller.

        Parameters
        ----------
        message : str
            Logging message
        level : logging.Level, optional
            Level of the logging message.
        """
        self.logger.log(message, level)

    @QtCore.pyqtSlot(int, int)
    def display_raw_data(self, timedelay_index, scan):
        """
        Extract raw data from raw dataset.

        Parameters
        ----------
        timedelay_index : int
            Time-delay index.
        scan : int
            Scan number.
        """
        timedelay = self.raw_dataset.time_points[timedelay_index]
        self.raw_data_signal.emit(self.raw_dataset.raw_data(timedelay, scan, bgr=True))
        self.status_message_signal.emit(
            f"Displaying data at {timedelay:.3f}ps, scan {scan:d}."
        )

    @QtCore.pyqtSlot(int)
    @QtCore.pyqtSlot(int, bool)
    def display_averaged_data(self, timedelay_index, autocontrast=False):
        """
        Extract processed diffraction pattern.

        Parameters
        ----------
        timedelay_index : int
            Time-delay index.
        autocontrast : bool, optional
            Whether or not the contrast of the image should be adjusted.

        Returns
        -------
        arr : `~numpy.ndarray`, ndim 2
            Diffracted intensity
        """
        # Preallocation of full images is important because the whole block cannot be
        # loaded into memory, contrary to powder data
        # Note that the containers have been initialized when the dataset was loaded
        # Therefore, no error checking required.
        self._timedelay_index = timedelay_index
        timedelay = self.dataset.time_points[timedelay_index]
        self._averaged_data_container[:] = self.dataset.diff_data(
            timedelay,
            relative=self._relative_averaged,
            out=self._averaged_data_container,
        )
        if self._relative_averaged:
            # The range of relative values can be outrageous if very large floats
            # are involved. This causes problems with pyqtgraph displays.
            # I doubt relative changes can exceed 50000%
            np.clip(
                self._averaged_data_container,
                a_min=-60000,
                a_max=60000,
                out=self._averaged_data_container,
            )

        self.averaged_data_signal.emit(self._averaged_data_container, autocontrast)
        self.status_message_signal.emit(f"Displaying data at {timedelay:.3f}ps.")

    @QtCore.pyqtSlot()
    def display_powder_data(self):
        """ Emit a powder data signal with/out background """
        # Preallocation isn't so important for powder data because the whole block
        # is loaded
        self.powder_data_signal.emit(
            self.dataset.scattering_vector,
            self.dataset.powder_data(
                timedelay=None, relative=self._relative_powder, bgr=self._bgr_powder
            ),
        )

    @QtCore.pyqtSlot(bool)
    def powder_background_subtracted(self, enable):
        """
        Toggle background subtraction for polycrystalline data.

        Parameters
        ----------
        enable : bool
        """
        self._bgr_powder = enable
        self.powder_bgr_enable_signal.emit(enable)
        self.display_powder_data()

    @QtCore.pyqtSlot(bool)
    def enable_powder_relative(self, enable):
        """
        Toggle relative scaling for polycrystalline data.

        Parameters
        ----------
        enable : bool
        """
        self._relative_powder = enable
        self.relative_powder_enable_signal.emit(enable)
        self.display_powder_data()

    @QtCore.pyqtSlot(bool)
    def enable_averaged_relative(self, enable):
        """
        Toggle relative scaling for diffracted data.

        Parameters
        ----------
        enable : bool
        """
        self._relative_averaged = enable
        self.relative_averaged_enable_signal.emit(enable)
        self.display_averaged_data(self._timedelay_index)

    @QtCore.pyqtSlot(float, float)
    def powder_time_series(self, qmin, qmax):
        """
        Extract polycrystalline time-series

        Parameters
        ----------
        qmin, qmax : float
            Lower- and upper-bounds of integration.
        """
        time_series = self.dataset.powder_time_series(
            rmin=qmin, rmax=qmax, bgr=self._bgr_powder, units="momentum"
        )
        self.powder_time_series_signal.emit(self.dataset.time_points, time_series)

    @QtCore.pyqtSlot(tuple)
    def time_series(self, rect):
        """ "
        Single-crystal time-series as the integrated diffracted intensity inside a rectangular ROI

        Parameters
        ----------
        rect : pyqtgraph.ROI
            Rectangle ROI defining integration bounds.
        """
        # Remember for updates
        self._rect = rect

        # Note that the containers have been initialized when the dataset was loaded
        # Therefore, no error checking required.
        integrated = self.dataset.time_series(
            rect, out=self._average_time_series_container
        )
        self.time_series_signal.emit(
            self.dataset.time_points, self._average_time_series_container
        )

    @QtCore.pyqtSlot(dict)
    def powder_calq(self, params):
        """
        Calibrate the range of q-vector for a polycrystalline data of known structure.

        Parameters
        ----------
        params : dict
            Parameters are passed to `skued.powder_calq`.
        """
        self.dataset.powder_calq(**params)
        self.display_powder_data()
        self.status_message_signal.emit("Scattering vector range calibrated.")

    @QtCore.pyqtSlot()
    @indicate_in_progress
    def find_bragg_peaks(self):
        """
        Determine the location of Bragg peaks for the equilibrium pattern
        """
        self.log("Finding Bragg peaks...", level=logging.DEBUG)
        peaks = bragg_peaks(
            im=self.dataset.diff_eq(),
            mask=self.dataset.valid_mask,
            center=self.dataset.center,
            min_dist=50,
        )
        self.bragg_peaks_signal.emit(peaks)

    @QtCore.pyqtSlot(dict)
    def update_metadata(self, metadata):
        """
        Update metadata attributes in DiffractionDataset

        Parameters
        ----------
        metadata : dict
            Metadata items to update. Metadata items not in ``DiffractionDataset.valid_metadata`` are ignored.
        """
        if self.dataset is None:
            return

        # dataset is read-only at this point
        # We close and reopen with write intentions
        fname = self.dataset.filename
        cls = type(self.dataset)
        self.dataset.close()

        with cls(fname, mode="r+") as dset:
            for key, val in metadata.items():
                if key in dset.valid_metadata:
                    setattr(dset, key, val)

        self.dataset = cls(fname, mode="r")
        self.dataset_metadata.emit(self.dataset.metadata)

        self.status_message_signal.emit("Metadata updated.")

    @QtCore.pyqtSlot(str)
    def set_dataset_notes(self, notes):
        """
        Update notes metadata in DiffractionDataset

        Parameters
        ----------
        notes : str
        """
        metadata = self.dataset.metadata
        metadata["notes"] = notes
        self.update_metadata(metadata)

    @QtCore.pyqtSlot(float)
    def set_time_zero_shift(self, shift):
        """
        Set the time-zero shift in picoseconds.

        Parameters
        ----------
        shift : float
            Time-delay shift in picoseconds.
        """
        if shift == self.dataset.time_zero_shift:
            return

        # dataset is read-only at this point
        # We close and reopen with write intentions
        # Note : no need to open as a PowderDiffractionDataset for this operations
        fname = self.dataset.filename
        cls = type(self.dataset)
        self.dataset.close()

        with DiffractionDataset(fname, mode="r+") as dset:
            dset.shift_time_zero(shift)

        self.dataset = cls(fname, mode="r")
        self.dataset_metadata.emit(self.dataset.metadata)

        # In case of a time-zero shift, diffraction time-series will
        # be impacted
        self.time_series(self._rect)

        # If _powder_relative is True, the shift in time-zero will impact the display
        if self._relative_powder:
            self.display_powder_data()

        self.status_message_signal.emit(f"Time-zero shifted by {shift:.3f}ps.")

    @QtCore.pyqtSlot(str, object)
    def load_raw_dataset(self, path, cls):
        """
        Load a raw dataset according to an AbstractRawDataset.

        Parameters
        ----------
        path : str
            Absolute path to the dataset.
        cls : AbstractRawDataset subclass
            Class to use during the loading.
        """
        if cls not in AbstractRawDataset.implementations:
            raise ValueError(
                f"Expected a proper subclass of AbstractRawDataset, but received {cls}"
            )

        if not path:
            return

        self.close_raw_dataset()
        self.raw_dataset = cls(path)
        self.raw_dataset_loaded_signal.emit(True)
        self.raw_dataset_metadata.emit(
            {
                "time_points": self.raw_dataset.time_points,
                "scans": self.raw_dataset.scans,
            }
        )
        self.display_raw_data(timedelay_index=0, scan=min(self.raw_dataset.scans))
        self.status_message_signal.emit(path + " loaded.")

    @QtCore.pyqtSlot()
    def close_raw_dataset(self):
        """ Close raw dataset. """
        self.raw_dataset = None
        self.raw_dataset_loaded_signal.emit(False)
        self.raw_data_signal.emit(None)

        self.status_message_signal.emit("Raw dataset closed.")

    @QtCore.pyqtSlot(str)
    def load_dataset(self, path):
        """
        Load dataset, distinguishing between PowderDiffractionDataset and DiffractionDataset.

        Parameters
        ----------
        path : str or path-like
            Path to the dataset.
        """
        if not path:  # e.g. path = ''
            return

        self.close_dataset()

        # First, open the dataset as if it was the base class
        # and perform migration if required
        with DiffractionDataset(path, mode="r"):
            self.logger.debug(f"Checking if {path} requires migration...")
        self.logger.debug(f"Migration check complete.")

        cls = DiffractionDataset
        with DiffractionDataset(path, mode="r") as d:
            if PowderDiffractionDataset._powder_group_name in d:
                cls = PowderDiffractionDataset
                is_powder = True
            else:
                is_powder = False

        # For powder datasets, there might be the creation of placeholder datasets
        # therefore, we open the file and do nothing
        # TODO: stop writing datasets in PowderDiffractionDataset.__init__
        #       since the GUI will only open datasets in read-mode by default
        if cls is PowderDiffractionDataset:
            with cls(path, mode="r+"):
                pass

        self.dataset = cls(path, mode="r")
        self.dataset_metadata.emit(self.dataset.metadata)

        # Initialize containers
        # This *must* be done before data is displayed
        self._averaged_data_container = np.empty(
            shape=self.dataset.resolution,
            dtype=self.dataset.diffraction_group["intensity"].dtype,
        )
        self._average_time_series_container = np.empty(
            shape=self.dataset.time_points.shape, dtype=float
        )

        self.processed_dataset_loaded_signal.emit(True)
        self.powder_dataset_loaded_signal.emit(is_powder)

        self.display_averaged_data(timedelay_index=0, autocontrast=True)
        if is_powder:
            self.display_powder_data()
        else:
            # Re-assert that a processed dataset was loaded
            # hence, switch views to the relevant widgets
            self.processed_dataset_loaded_signal.emit(True)

        self.status_message_signal.emit(path + " loaded.")

    @QtCore.pyqtSlot()
    def close_dataset(self):
        """ Close current DiffractionDataset. """
        with suppress(AttributeError):  # in case self.dataset is None
            self.dataset.close()
        self.dataset = None
        self.processed_dataset_loaded_signal.emit(False)
        self.powder_dataset_loaded_signal.emit(False)

        self.averaged_data_signal.emit(None, True)
        self.powder_data_signal.emit(None, None)

        self.status_message_signal.emit("Dataset closed.")

    @QtCore.pyqtSlot(str, dict)
    def symmetrize(self, destination, params):
        """
        Launches a background thread that copies the currently-loaded dataset,
        symmetrize the copy, and load the copy.

        Parameters
        ----------
        destination : str or path-like
            Destination of the symmetrized dataset.
        params : dict
            Symmetrization parameters are passed to ``DiffractionDataset.symmetrize``.
        """
        kwargs = {
            "dataset": self.dataset.filename,
            "destination": destination,
            "callback": self.processing_progress_signal.emit,
        }
        kwargs.update(params)

        # TODO:
        #   Since symmetrization will first copy the dataset,
        #   and the dataset is open in read-only mode, do we have to close it?
        self.close_dataset()

        self.worker = WorkThread(function=symmetrize, kwargs=kwargs)
        self.worker.results_signal.connect(self.load_dataset)
        self.worker.in_progress_signal.connect(self.processing_data_signal)
        self.worker.in_progress_signal.connect(self.operation_in_progress)
        self.worker.done_signal.connect(
            lambda: self.processing_progress_signal.emit(100)
        )
        self.processing_progress_signal.emit(0)
        self.worker.start()

    @QtCore.pyqtSlot(dict)
    def process_raw_dataset(self, info_dict):
        """
        Launch a background thread that reduces raw data.

        Parameters
        ----------
        info_dict : dict
            Data reduction parameters are passed to ``DiffractionDataset.from_raw``
        """
        info_dict.update(
            {"callback": self.processing_progress_signal.emit, "raw": self.raw_dataset}
        )

        self.worker = WorkThread(function=process, kwargs=info_dict)
        self.worker.results_signal.connect(self.load_dataset)
        self.worker.in_progress_signal.connect(self.processing_data_signal)
        self.worker.in_progress_signal.connect(self.operation_in_progress)
        self.worker.done_signal.connect(
            lambda: self.processing_progress_signal.emit(100)
        )
        self.processing_progress_signal.emit(0)
        self.worker.start()

    @QtCore.pyqtSlot(dict)
    def calculate_azimuthal_averages(self, params):
        """
        Promote a DiffractionDataset to a PowderDiffractionDataset

        Parameters
        ----------
        params : dict
            Parameters are passed to ``PowderDiffractionDataset.from_dataset``.
        """
        params.update(
            {
                "filename": self.dataset.filename,
                "callback": self.powder_promotion_progress.emit,
            }
        )
        self.worker = WorkThread(function=calculate_azimuthal_averages, kwargs=params)

        self.close_dataset()

        self.worker.results_signal.connect(self.load_dataset)
        self.worker.in_progress_signal.connect(self.processing_data_signal)
        self.worker.in_progress_signal.connect(self.operation_in_progress)
        self.worker.start()

    @QtCore.pyqtSlot(dict)
    def compute_baseline(self, params):
        """Compute the powder baseline. The dictionary `params` is passed to
        PowderDiffractionDataset.compute_baseline(), except its key 'callback'. The callable 'callback'
        is called (no argument) when computation is done."""
        params.update({"fname": self.dataset.filename})

        self.worker = WorkThread(function=compute_powder_baseline, kwargs=params)
        self.dataset.close()
        self.dataset = None

        self.worker.results_signal.connect(self.load_dataset)
        self.worker.done_signal.connect(lambda: self.powder_background_subtracted(True))
        self.worker.in_progress_signal.connect(self.operation_in_progress)
        self.worker.start()


class WorkThread(QtCore.QThread):
    """
    Object taking care of threading computations. These computations are very specific:
    a function takes in a filename (pointing to a DiffractionDataset), parameters,
    and returns a filename again.

    Signals
    -------
    done_signal
        Emitted when the function evaluation is over.
    in_progress_signal
        Emitted when the function evaluation starts.
    results_signal
        Emitted when the function evaluation is over. Carries the results
        of the computation.
    """

    results_signal = QtCore.pyqtSignal(str)
    done_signal = QtCore.pyqtSignal()
    in_progress_signal = QtCore.pyqtSignal(bool)

    def __init__(self, function, args=tuple(), kwargs=dict()):

        QtCore.QThread.__init__(self)
        self.function = function
        self.args = args
        self.kwargs = kwargs

    def __del__(self):
        self.wait()

    def run(self):
        self.in_progress_signal.emit(True)

        # This is potentially a very-long-running calculation
        result = self.function(*self.args, **self.kwargs)

        self.results_signal.emit(result)
        self.done_signal.emit()
        self.in_progress_signal.emit(False)


def calculate_azimuthal_averages(**kwargs):
    """Create a PowderDiffractionDataset from a DiffractionDataset. If azimuthal averages
    were already calculated, recalculate them."""
    filename = kwargs.pop("filename")

    # Determine if azimuthal averages have already been computed
    recomputed = False
    with PowderDiffractionDataset(filename) as d:
        if PowderDiffractionDataset._powder_group_name in d:
            # We simply need to recompute the azimuthal averages
            recomputed = True
            d.compute_angular_averages(**kwargs)

    if recomputed:
        return filename

    # If we are here, the dataset neets promotion
    with PowderDiffractionDataset.from_dataset(DiffractionDataset(filename), **kwargs):
        pass

    return filename


def symmetrize(dataset, destination, **kwargs):
    """
    Copies a dataset and symmetrize it. Keyword arguments
    are passed to `DiffractionDataset.symmetrize`.

    Parameters
    ----------
    dataset : path-like
    destination : path-like
    """
    copy2(dataset, destination)

    with DiffractionDataset(destination, mode="r+") as dset:
        dset.symmetrize(**kwargs)
        fname = dset.filename
    return fname


def compute_powder_baseline(fname, **kwargs):
    """
    Compute a powder baseline. Keyword arguments are passed to
    `PowderDiffractionDataset.compute_baseline`

    Parameters
    ----------
    dataset : path-like
    """
    with PowderDiffractionDataset(fname, mode="r+") as dset:
        dset.compute_baseline(**kwargs)
    return fname


def process(**kwargs):
    """ Process a RawDataset into a DiffractionDataset """
    with DiffractionDataset.from_raw(**kwargs) as dset:
        fname = dset.filename
    return fname



# -*- coding: utf-8 -*-

import logging
import os
import platform
import sys
from contextlib import contextmanager
from os.path import join
from pathlib import Path
from subprocess import Popen
from warnings import warn

import pyqtgraph as pg
from PyQt5 import QtGui

from ..raw import open_raw
from .gui import Iris, IMAGE_FOLDER
from qdarkstyle import load_stylesheet_pyqt5


@contextmanager
def gui_environment():
    """
    Prepare the environment in which iris GUI will run. This includes the following:

        * Set the PyQtGraph QT library to PyQt5 while Iris GUI is running. Revert back when done.
        * Set the image-axis order to row-major. Revert back when done.

    Note that interactions with the screen (e.g. mask creation) assumes that the image-axis order is
    row-major.
    """
    old_qt_lib = os.environ.get(
        "PYQTGRAPH_QT_LIB", "PyQt5"
    )  # environment variable might not exist
    os.environ["PYQTGRAPH_QT_LIB"] = "PyQt5"

    old_image_axis_order = pg.getConfigOption("imageAxisOrder")
    pg.setConfigOptions(imageAxisOrder="row-major")

    yield
    os.environ["PYQTGRAPH_QT_LIB"] = old_qt_lib
    pg.setConfigOptions(imageAxisOrder=old_image_axis_order)


def run(path=None, dset_type=None, **kwargs):
    """
    Run the iris GUI with the correct environment, and open a dataset. Invalid
    datasets are ignored.

    Parameters
    ----------
    path : path-like or None, optional
        Path to either a raw dataset or a processed datasets.
        Raw dataset formats will be guessed.
    dset_type : {'raw', 'reduced', None}, optional
        Dataset type.
    """

    with gui_environment():
        app = QtGui.QApplication(sys.argv)
        app.setStyleSheet(load_stylesheet_pyqt5())
        app.setWindowIcon(QtGui.QIcon(join(IMAGE_FOLDER, "eye.png")))
        gui = Iris()

        if path:
            path = Path(path)
            if dset_type == "raw":
                # Determine the class
                try:
                    with open_raw(path) as dset:
                        dataformat = dset.__class__
                except RuntimeError:
                    pass
                else:
                    # No errors, valid dataset
                    # note : signal has signature [str, object]
                    gui.raw_dataset_path_signal.emit(str(path), dataformat)
            if dset_type == "reduced":
                gui.dataset_path_signal.emit(str(path))  # signal has signature [str]
            else:
                warn(f"dset_type invalid value: {dset_type}. Ignoring path.")

        result = app.exec_()
        logging.shutdown()
        return result



import sys
import os
from contextlib import suppress
from itertools import repeat
from tempfile import gettempdir

import numpy as np
from numpy.random import random

from crystals import Crystal
from iris import DiffractionDataset
from iris.dataset import SWMR_AVAILABLE
from skued import nfold, RectSelection, DiskSelection, RingSelection, ArbitrarySelection
from pathlib import Path
from . import TestRawDataset
import pytest

np.random.seed(23)


def double(im):
    return im * 2


@pytest.fixture
def fname():
    f = Path(gettempdir()) / "test.hdf5"
    yield f
    with suppress(OSError):
        os.remove(f)


def test_creation_from_raw_default(fname):
    """ Test that DiffractionDataset.from_raw() works with default settigns """
    raw = TestRawDataset()

    with DiffractionDataset.from_raw(raw, filename=fname, mode="w") as dataset:
        assert dataset.diffraction_group["intensity"].shape == raw.resolution + (
            len(raw.time_points),
        )


def test_creation_from_raw_alignment(fname):
    """ Test that DiffractionDataset.from_raw(..., align = True) does not throw any errors """
    raw = TestRawDataset()

    with DiffractionDataset.from_raw(
        raw, filename=fname, align=True, mode="w"
    ) as dataset:
        assert dataset.diffraction_group["intensity"].shape == raw.resolution + (
            len(raw.time_points),
        )


def test_creation_from_raw_multiprocess(fname):
    """ Test that DiffractionDataset.from_raw(..., processes = 2) does not throw any errors """
    raw = TestRawDataset()

    with DiffractionDataset.from_raw(
        raw, filename=fname, align=False, processes=2, mode="w"
    ) as dataset:
        assert dataset.diffraction_group["intensity"].shape == raw.resolution + (
            len(raw.time_points),
        )


def test_creation_from_collection(fname):
    """ Test the creation of a DiffractionDataset from a collection of patterns """
    patterns = repeat(random(size=(256, 256)), 10)
    metadata = {"fluence": 10, "energy": 90}

    with DiffractionDataset.from_collection(
        patterns,
        filename=fname,
        time_points=list(range(10)),
        metadata=metadata,
        dtype=np.float16,
        mode="w",
    ) as dataset:

        assert dataset.diffraction_group["intensity"].shape == (256, 256, 10)
        assert dataset.diffraction_group["intensity"].dtype == np.float16
        assert dataset.fluence == metadata["fluence"]
        assert dataset.energy == metadata["energy"]
        assert list(dataset.time_points) == list(map(float, range(10)))


@pytest.fixture
def dataset():
    patterns = list(repeat(random(size=(256, 256)), 5))
    metadata = {"fluence": 10, "energy": 90}
    filename = Path(gettempdir()) / "test.hdf5"
    dset = DiffractionDataset.from_collection(
        patterns,
        filename=filename,
        time_points=range(5),
        metadata=metadata,
        mode="w",
    )
    setattr(dset, "patterns", patterns)
    yield dset
    dset.close()
    del dset
    with suppress(OSError):
        os.remove(filename)


def test_file_modes(dataset):
    """ Successively open and close the same dataset with different file modes. """
    fname = dataset.filename
    metadata = dataset.metadata
    dataset.close()

    with DiffractionDataset(fname, mode="r") as dset:
        assert metadata == dset.metadata
        assert dset.mode == "r"

    for mode in ("r+", "a"):
        with DiffractionDataset(fname, mode=mode) as dset:
            assert metadata == dset.metadata
            assert dset.mode == "r+"

    with pytest.raises(OSError):
        DiffractionDataset(fname, mode="x")

    # Reopen dataset so it can be deleted
    dataset = DiffractionDataset(fname, mode="r")


def test_write_access(dataset):
    """ Check that certain operations respect write access """
    fname = dataset.filename
    metadata = dataset.metadata
    dataset.close()

    with DiffractionDataset(fname, mode="r") as dset:
        with pytest.raises(IOError):
            dset.symmetrize(mod=3, center=(0, 0))

    # Reopen dataset so it can be deleted
    dataset = DiffractionDataset(fname, mode="r")


def test_dataset_metadata(dataset):
    """ Test that the property 'metadata' is working correctly"""
    metadata = dataset.metadata
    for required in DiffractionDataset.valid_metadata:
        assert required in metadata
    assert "filename" in metadata


def test_notes(dataset):
    """ Test that updating the notes works as intended """
    dataset.notes = "test notes"
    assert dataset.notes == "test notes"
    dataset.notes = "different notes"
    assert dataset.notes == "different notes"


def test_diff_apply(dataset):
    """ Test that the diff_apply method works as expected """
    before = np.array(dataset.diffraction_group["intensity"])
    dataset.diff_apply(lambda arr: arr * 2)
    after = np.array(dataset.diffraction_group["intensity"])
    assert np.allclose(2 * before, after)

    with pytest.raises(TypeError):
        dataset.diff_apply(None)


@pytest.mark.skipif(not SWMR_AVAILABLE, reason="Parallel execution is not available")
def test_diff_apply_parallel(dataset):
    """ Test that the diff_apply method works as expected in parallel mode """
    before = np.array(dataset.diffraction_group["intensity"])
    dataset.diff_apply(double, processes=2)
    after = np.array(dataset.diffraction_group["intensity"])
    assert np.allclose(2 * before, after)

    with pytest.raises(TypeError):
        dataset.diff_apply(None)


def test_mask_apply(dataset):
    """ test that DiffractionDataset.mask_apply method works as expected """
    old_mask = dataset.valid_mask
    random_mask = np.random.random(size=old_mask.shape) > 0.5
    func = lambda m: random_mask
    new_mask = func(old_mask)
    dataset.mask_apply(func)
    assert np.allclose(dataset.valid_mask, new_mask)

    with pytest.raises(TypeError):
        dataset.mask_apply(None)

    func2 = lambda m: m.astype(float)
    with pytest.raises(TypeError):
        dataset.mask_apply(func2)

    func3 = lambda m: np.zeros((m.shape[0] // 2, m.shape[1] // 2), dtype=bool)
    with pytest.raises(ValueError):
        dataset.mask_apply(func3)


def test_symmetrization(dataset):
    """ Test correctness of symmetrization operation """
    before = np.array(dataset.diffraction_group["intensity"])
    symmetrized = np.array(before, copy=True)
    for index, _ in enumerate(dataset.time_points):
        symmetrized[:, :, index] = nfold(
            before[:, :, index], mod=3, center=(63, 65), mask=dataset.valid_mask
        )

    dataset.symmetrize(mod=3, center=(63, 65))
    after = np.array(dataset.diffraction_group["intensity"])

    assert np.allclose(symmetrized, after)


@pytest.mark.skipif(not SWMR_AVAILABLE, reason="Parallel execution is not available")
def test_symmetrization_parallel(dataset):
    """ Test correctness of symmetrization operation in parallel mode """
    before = np.array(dataset.diffraction_group["intensity"])
    symmetrized = np.array(before, copy=True)
    for index, _ in enumerate(dataset.time_points):
        symmetrized[:, :, index] = nfold(
            before[:, :, index], mod=3, center=(63, 65), mask=dataset.valid_mask
        )

    dataset.symmetrize(mod=3, center=(63, 65), processes=2)
    after = np.array(dataset.diffraction_group["intensity"])

    assert np.allclose(symmetrized, after)


def test_data(dataset):
    """ Test that data stored in DiffractionDataset is correct """
    for time, pattern in zip(list(dataset.time_points), getattr(dataset, "patterns")):
        assert np.allclose(dataset.diff_data(time), pattern)


def test_time_zero_shift(dataset):
    """ Test that data changed with time_zero_shift() """
    unshifted = np.array(dataset.time_points)
    dataset.shift_time_zero(100)
    shifted = np.array(dataset.time_points)

    assert not np.allclose(unshifted, shifted)
    assert np.allclose(unshifted + 100, shifted)


def test_diff_eq(dataset):
    """ test that DiffractionDataset.diff_eq() returns the correct array """
    dataset.shift_time_zero(10)
    assert np.allclose(dataset.diff_eq(), np.zeros(dataset.resolution))

    dataset.shift_time_zero(-20)
    eq = np.mean(np.stack(getattr(dataset, "patterns"), axis=-1), axis=2)
    assert np.allclose(dataset.diff_eq(), eq)


def test_time_series(dataset):
    """ Test that the DiffractionDataset.time_series method is working """

    r1, r2, c1, c2 = 100, 120, 45, 57
    stack = np.stack(getattr(dataset, "patterns"), axis=-1)
    ts = np.mean(stack[r1:r2, c1:c2], axis=(0, 1))

    assert np.allclose(dataset.time_series([r1, r2, c1, c2], relative=False), ts)

    assert np.allclose(dataset.time_series([r1, r2, c1, c2], relative=True), ts)


def test_time_series_selection(dataset):
    """Test that the DiffractionDataset.time_series_selection
    method is working as expected"""
    mask = np.random.choice([True, False], size=dataset.resolution)
    selection = ArbitrarySelection(mask)

    stack = np.stack(getattr(dataset, "patterns"), axis=-1)
    ts = np.mean(stack[selection], axis=(0, 1))

    assert np.allclose(dataset.time_series_selection(selection, relative=False), ts)

    assert np.allclose(dataset.time_series_selection(selection, relative=True), ts)


def test_selection_rect(dataset):
    """Comparison of DiffractionDataset.time_series vs
    DiffractionDataset.time_series_selection with
    DiffractionDataset.selection_rect"""

    r1, r2, c1, c2 = 100, 120, 45, 57
    selection = RectSelection(dataset.resolution, r1, r2, c1, c2)

    ts = dataset.time_series([r1, r2 + 1, c1, c2 + 1], relative=False)
    tsbm = dataset.time_series_selection(selection, relative=False)
    assert np.allclose(ts, tsbm)

    ts = dataset.time_series([r1, r2 + 1, c1, c2 + 1], relative=True)
    tsbm = dataset.time_series_selection(selection, relative=True)
    assert np.allclose(ts, tsbm)


def test_selection_disk_relative(dataset):
    """Test relative DiffractionDataset.time_series_selection with DiffractionDataset.selection_disk"""
    selection = DiskSelection(dataset.resolution, center=(120, 200), radius=10)

    # We modify the dataset so that within the selection, only zeroes are found
    # Note that HDF5 does not support fancy boolean indexing, so we must change the
    # content image-by-image.
    for index, _ in enumerate(dataset.time_points):
        arr = dataset.diffraction_group["intensity"][:, :, index]
        arr[selection] = 0
        dataset.diffraction_group["intensity"][:, :, index] = arr

    ts = dataset.time_series_selection(selection, relative=False)
    assert np.allclose(ts, np.zeros_like(ts))


def test_selection_disk_nonrelative(dataset):
    """Test DiffractionDataset.time_series_selection with DiffractionDataset.selection_disk"""
    selection = DiskSelection(dataset.resolution, center=(120, 200), radius=10)

    for index, _ in enumerate(dataset.time_points):
        arr = dataset.diffraction_group["intensity"][:, :, index]
        arr[selection] = 1
        dataset.diffraction_group["intensity"][:, :, index] = arr

    ts = dataset.time_series_selection(selection, relative=True)
    assert np.allclose(ts, np.ones_like(ts))


def test_selection_ring(dataset):
    """Test DiffractionDataset.time_series_selection with DiffractionDataset.selection_ring"""
    selection = RingSelection(
        dataset.resolution, center=(120, 200), inner_radius=10, outer_radius=20
    )

    # We modify the dataset so that within the selection, only zeroes are found
    # Note that HDF5 does not support fancy boolean indexing, so we must change the
    # content image-by-image.
    for index, _ in enumerate(dataset.time_points):
        arr = dataset.diffraction_group["intensity"][:, :, index]
        arr[selection] = 0
        dataset.diffraction_group["intensity"][:, :, index] = arr

    ts = dataset.time_series_selection(selection, relative=False)
    assert np.allclose(ts, np.zeros_like(ts))


def test_selection_ring_relative(dataset):
    """Test relative DiffractionDataset.time_series_selection with DiffractionDataset.selection_ring"""
    selection = RingSelection(
        dataset.resolution, center=(120, 200), inner_radius=10, outer_radius=20
    )

    for index, _ in enumerate(dataset.time_points):
        arr = dataset.diffraction_group["intensity"][:, :, index]
        arr[selection] = 1
        dataset.diffraction_group["intensity"][:, :, index] = arr

    ts = dataset.time_series_selection(selection, relative=True)
    assert np.allclose(ts, np.ones_like(ts))



from functools import lru_cache, partial, wraps
from math import sqrt

import numpy as np

from npstreams import peek, pmap
from skued import (
    __version__,
    azimuthal_average,
    baseline_dt,
    autocenter,
    powder_calq,
    ArbitrarySelection,
    Selection,
)
from skued.baseline import dt_max_level

from .meta import HDF5ExperimentalParameter, MetaHDF5Dataset

from .dataset import DiffractionDataset, write_access_needed


class PowderDiffractionDataset(DiffractionDataset):
    """
    Abstraction of HDF5 files for powder diffraction datasets.
    """

    _powder_group_name = "/powder"

    angular_bounds = HDF5ExperimentalParameter(
        "angular_bounds", tuple, default=(0, 360)
    )
    first_stage = HDF5ExperimentalParameter(
        "powder_baseline_first_stage", str, default=""
    )
    wavelet = HDF5ExperimentalParameter("powder_baseline_wavelet", str, default="")
    level = HDF5ExperimentalParameter("powder_baseline_level", int, default=0)
    niter = HDF5ExperimentalParameter("powder_baseline_niter", int, default=0)

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        # TODO: stop writing datasets in __init__
        #       since the GUI will only open datasets in read-mode by default

        # Ensure that all required powder groups exist
        maxshape = (len(self.time_points), sqrt(2 * max(self.resolution) ** 2))
        for name in {"intensity", "baseline"}:
            if name not in self.powder_group:
                self.powder_group.create_dataset(
                    name=name,
                    shape=maxshape,
                    maxshape=maxshape,
                    dtype=float,
                    fillvalue=0.0,
                    **self.compression_params,
                )

        # Radius from center in units of pixels
        shape = self.powder_group["intensity"].shape
        placeholder = np.arange(0, shape[-1])
        if "px_radius" not in self.powder_group:
            self.powder_group.create_dataset(
                "px_radius", data=placeholder, maxshape=(maxshape[-1],), dtype=float
            )

        # Radius from center in units of inverse angstroms
        if "scattering_vector" not in self.powder_group:
            self.powder_group.create_dataset(
                "scattering_vector",
                data=placeholder,
                maxshape=(maxshape[-1],),
                dtype=float,
            )

    @classmethod
    def from_dataset(
        cls, dataset, center=None, normalized=True, angular_bounds=None, callback=None
    ):
        """
        Transform a DiffractionDataset instance into a PowderDiffractionDataset. This requires
        computing the azimuthal averages as well.

        Parameters
        ----------
        dataset : DiffractionDataset
            DiffractionDataset instance.
        center : 2-tuple or None, optional
            Center of the diffraction patterns. If None (default), center will be automatically-determined.
        normalized : bool, optional
            If True, each pattern is normalized to its integral. Default is False.
        angular_bounds : 2-tuple of float or None, optional
            Angle bounds are specified in degrees. 0 degrees is defined as the positive x-axis.
            Angle bounds outside [0, 360) are mapped back to [0, 360).
        callback : callable or None, optional
            Callable of a single argument, to which the calculation progress will be passed as
            an integer between 0 and 100.

        Returns
        -------
        powder : PowderDiffractionDataset
        """
        fname = dataset.filename
        dataset.close()

        powder_dataset = cls(fname, mode="r+")
        powder_dataset.compute_angular_averages(
            center, normalized, angular_bounds, callback
        )
        return powder_dataset

    @property
    def powder_group(self):
        return self.require_group(self._powder_group_name)

    @property
    def px_radius(self):
        """ Pixel-radius of azimuthal average """
        return np.array(self.powder_group["px_radius"])

    @property
    def scattering_vector(self):
        """ Array of scattering vector norm :math:`|q|` [:math:`1/\\AA`] """
        return np.array(self.powder_group["scattering_vector"])

    def shift_time_zero(self, *args, **kwargs):
        """
        Shift time-zero uniformly across time-points.

        Parameters
        ----------
        shift : float
            Shift [ps]. A positive value of `shift` will move all time-points forward in time,
            whereas a negative value of `shift` will move all time-points backwards in time.
        """
        self.powder_eq.cache_clear()
        return super().shift_time_zero(*args, **kwargs)

    @write_access_needed
    def powder_calq(self, crystal, peak_indices, miller_indices):
        """
        Determine the scattering vector q corresponding to a polycrystalline diffraction pattern
        and a known crystal structure.

        For best results, multiple peaks (and corresponding Miller indices) should be provided; the
        absolute minimum is two.

        Parameters
        ----------
        crystal : skued.Crystal instance
            Crystal that gave rise to the diffraction data.
        peak_indices : n-tuple of ints
            Array index location of diffraction peaks. For best
            results, peaks should be well-separated. More than two peaks can be used.
        miller_indices : iterable of 3-tuples
            Indices associated with the peaks of ``peak_indices``. More than two peaks can be used.
            E.g. ``indices = [(2,2,0), (-3,0,2)]``

        Raises
        ------
        ValueError
            if the number of peak indices does not match the number of Miller indices, or if the
            number of peaks given is lower than two.
        IOError
            If the filename is already associated with a file.
        """
        I = self.powder_eq()
        q = powder_calq(
            I=I,
            crystal=crystal,
            peak_indices=peak_indices,
            miller_indices=miller_indices,
        )

        self.powder_group["scattering_vector"].resize(I.shape)
        self.powder_group["scattering_vector"].write_direct(q)

    @lru_cache(maxsize=2)  # with and without background
    def powder_eq(self, bgr=False):
        """
        Returns the average powder diffraction pattern for all times before photoexcitation.
        In case no data is available before photoexcitation, an array of zeros is returned.

        Parameters
        ----------
        bgr : bool
            If True, background is removed.

        Returns
        -------
        I : ndarray, shape (N,)
            Diffracted intensity [counts]
        """
        t0_index = np.argmin(np.abs(self.time_points))
        b4t0_slice = self.powder_group["intensity"][:t0_index, :]

        # If there are no available data before time-zero, np.mean()
        # will return an array of NaNs; instead, return zeros.
        if t0_index == 0:
            return np.zeros_like(self.px_radius)

        if not bgr:
            return np.mean(b4t0_slice, axis=0)

        bg = self.powder_group["baseline"][:t0_index, :]
        return np.mean(b4t0_slice - bg, axis=0)

    def powder_data(self, timedelay, bgr=False, relative=False, out=None):
        """
        Returns the angular average data from scan-averaged diffraction patterns.

        Parameters
        ----------
        timdelay : float or None
            Time-delay [ps]. If None, the entire block is returned.
        bgr : bool, optional
            If True, background is removed.
        relative : bool, optional
            If True, data is returned relative to the average of all diffraction patterns
            before photoexcitation.
        out : ndarray or None, optional
            If an out ndarray is provided, h5py can avoid
            making intermediate copies.

        Returns
        -------
        I : ndarray, shape (N,) or (N,M)
            Diffracted intensity [counts]
        """
        dataset = self.powder_group["intensity"]

        if timedelay is None:
            if out is None:
                out = np.empty_like(dataset)
            dataset.read_direct(out)

        else:
            time_index = self._get_time_index(timedelay)
            if out is None:
                out = np.empty_like(self.px_radius)
            dataset.read_direct(out, source_sel=np.s_[time_index, :])

        if bgr:
            out -= self.powder_baseline(timedelay)

        if relative:
            out -= self.powder_eq(bgr=bgr)

        return out

    def powder_baseline(self, timedelay, out=None):
        """
        Returns the baseline data.

        Parameters
        ----------
        timdelay : float or None
            Time-delay [ps]. If None, the entire block is returned.
        out : ndarray or None, optional
            If an out ndarray is provided, h5py can avoid
            making intermediate copies.

        Returns
        -------
        out : ndarray
            If a baseline hasn't been computed yet, the returned
            array is an array of zeros.
        """
        try:
            dataset = self.powder_group["baseline"]
        except KeyError:
            return np.zeros_like(self.px_radius)

        if timedelay is None:
            if out is None:
                out = np.empty_like(dataset)
            dataset.read_direct(out)

        else:
            time_index = self._get_time_index(timedelay)
            if out is None:
                out = np.empty_like(self.px_radius)
            dataset.read_direct(out, source_sel=np.s_[time_index, :])

        return out

    def powder_time_series(
        self, rmin, rmax, bgr=False, relative=False, units="pixels", out=None
    ):
        """
        Average intensity over time.
        Diffracted intensity is integrated in the closed interval [rmin, rmax]

        Parameters
        ----------
        rmin : float
            Lower scattering vector bound [1/A]
        rmax : float
            Higher scattering vector bound [1/A].
        bgr : bool, optional
            If True, background is removed. Default is False.
        relative : bool, optional
            If True, data is returned relative to the average of all diffraction patterns
            before photoexcitation.
        units : str, {'pixels', 'momentum'}
            Units of the bounds rmin and rmax.
        out : ndarray or None, optional
            1-D ndarray in which to store the results. The shape
            should be compatible with (len(time_points),)

        Returns
        -------
        out : ndarray, shape (N,)
            Average diffracted intensity over time.
        """
        # In some cases, it is easier
        if units not in {"pixels", "momentum"}:
            raise ValueError(
                f"``units`` must be either 'pixels' or 'momentum', not {units}"
            )
        abscissa = self.px_radius if units == "pixels" else self.scattering_vector

        i_min, i_max = (
            np.argmin(np.abs(rmin - abscissa)),
            np.argmin(np.abs(rmax - abscissa)),
        )
        i_max += 1  # Python slices are semi-open by design, therefore i_max + 1 is used
        trace = np.array(self.powder_group["intensity"][:, i_min:i_max])
        if bgr:
            trace -= np.array(self.powder_group["baseline"][:, i_min:i_max])

        if relative:
            trace -= self.powder_eq(bgr=bgr)[i_min:i_max]

        if out is not None:
            return np.mean(axis=1, out=out)
        return np.mean(trace, axis=1).reshape(-1)

    @write_access_needed
    def compute_baseline(self, first_stage, wavelet, max_iter=50, level=None, **kwargs):
        """
        Compute and save the baseline computed based on the dual-tree complex wavelet transform.
        All keyword arguments are passed to scikit-ued's `baseline_dt` function.

        Parameters
        ----------
        first_stage : str, optional
            Wavelet to use for the first stage. See :func:`skued.available_first_stage_filters` for a list of suitable arguments
        wavelet : str, optional
            Wavelet to use in stages > 1. Must be appropriate for the dual-tree complex wavelet transform.
            See :func:`skued.available_dt_filters` for possible values.
        max_iter : int, optional

        level : int or None, optional
            If None (default), maximum level is used.

        Raises
        ------
        IOError
            If the filename is already associated with a file.
        """
        block = self.powder_data(timedelay=None, bgr=False)

        baseline_kwargs = {
            "array": block,
            "max_iter": max_iter,
            "level": level,
            "first_stage": first_stage,
            "wavelet": wavelet,
            "axis": 1,
        }
        baseline_kwargs.update(**kwargs)

        baseline = np.ascontiguousarray(
            baseline_dt(**baseline_kwargs)
        )  # In rare cases this wasn't C-contiguous

        # The baseline dataset is guaranteed to exist after compte_angular_averages was called.
        self.powder_group["baseline"].resize(baseline.shape)
        self.powder_group["baseline"].write_direct(baseline)

        if level == None:
            level = dt_max_level(
                data=self.px_radius, first_stage=first_stage, wavelet=wavelet
            )

        self.level = level
        self.first_stage = first_stage
        self.wavelet = wavelet
        self.niter = max_iter

        self.powder_eq.cache_clear()

    @write_access_needed
    def compute_angular_averages(
        self,
        center=None,
        normalized=False,
        angular_bounds=None,
        trim=True,
        callback=None,
    ):
        """
        Compute the angular averages.

        Parameters
        ----------
        center : 2-tuple or None, optional
            Center of the diffraction patterns. If None (default), the dataset
            attribute will be used instead. If that is not possible, the center will
            be automatically determined. See :meth:`DiffractionDataset.autocenter`.
        normalized : bool, optional
            If True, each pattern is normalized to its integral.
        angular_bounds : 2-tuple of float or None, optional
            Angle bounds are specified in degrees. 0 degrees is defined as the positive x-axis.
            Angle bounds outside [0, 360) are mapped back to [0, 360).
        trim : bool, optional
            If True, leading/trailing zeros - possibly due to masks - are trimmed.
        callback : callable or None, optional
            Callable of a single argument, to which the calculation progress will be passed as
            an integer between 0 and 100.

        Raises
        ------
        IOError
            If the filename is already associated with a file.
        """
        # TODO: allow to cut away regions
        if not any([self.center, center]):
            raise RuntimeError(
                "Center attribute must be either saved in the dataset \
                                as an attribute or be provided."
            )

        if callback is None:
            callback = lambda i: None

        if center is not None:
            self.center = center

        if center is None:
            center = self.center

        # Because it is difficult to know the angular averaged data's shape in advance,
        # we calculate it first and store it next
        callback(0)
        results = list()
        for index, timedelay in enumerate(self.time_points):
            px_radius, avg = azimuthal_average(
                self.diff_data(timedelay),
                center=center,
                mask=self.valid_mask,
                angular_bounds=angular_bounds,
                trim=False,
            )

            # px_radius is not stored but used once
            results.append(avg)
            callback(int(100 * index / len(self.time_points)))

        # Concatenate arrays for intensity and error
        # If trimming is enabled, there might be a problem where
        # different averages are trimmed to different length
        # therefore, we trim to the most restrictive bounds
        if trim:
            bounds = [_trim_bounds(I) for I in results]
            min_bound = max(min(bound) for bound in bounds)
            max_bound = min(max(bound) for bound in bounds)
            results = [I[min_bound:max_bound] for I in results]
            px_radius = px_radius[min_bound:max_bound]

        rintensity = np.stack(results, axis=0)

        if normalized:
            rintensity /= np.sum(rintensity, axis=1, keepdims=True)

        # We allow resizing. In theory, an angular averave could never be
        # longer than the diagonal of resolution
        self.powder_group["intensity"].resize(rintensity.shape)
        self.powder_group["intensity"].write_direct(rintensity)

        self.powder_group["px_radius"].resize(px_radius.shape)
        self.powder_group["px_radius"].write_direct(px_radius)

        # Use px_radius as placeholder for scattering_vector until calibration
        self.powder_group["scattering_vector"].resize(px_radius.shape)
        self.powder_group["scattering_vector"].write_direct(px_radius)

        self.powder_group["baseline"].resize(rintensity.shape)
        self.powder_group["baseline"].write_direct(np.zeros_like(rintensity))

        self.powder_eq.cache_clear()
        callback(100)


def _trim_bounds(arr):
    """ Returns the bounds which would be used in numpy.trim_zeros but also trimmming nans"""
    first = 0
    for i in arr:
        if i != 0.0:
            break
        else:
            first = first + 1
    last = len(arr)
    for i in arr[::-1]:
        if i != 0.0:
            break
        else:
            last = last - 1
    return first, last



import sys
import os
from contextlib import suppress
from itertools import repeat
from tempfile import gettempdir

import numpy as np
from numpy.random import random

from crystals import Crystal
from iris import PowderDiffractionDataset, DiffractionDataset
from pathlib import Path
import pytest

np.random.seed(23)


@pytest.fixture
def powder_dataset():
    patterns = list(repeat(random(size=(128, 128)), 5))
    metadata = {"fluence": 10, "energy": 90}
    filename = Path(gettempdir()) / "test.hdf5"
    diff_dataset = DiffractionDataset.from_collection(
        patterns,
        filename=filename,
        time_points=range(5),
        metadata=metadata,
        mode="w",
    )
    dset = PowderDiffractionDataset.from_dataset(diff_dataset, center=(23, 45))
    setattr(dset, "patterns", patterns)
    yield dset
    dset.close()
    del dset
    with suppress(OSError):
        os.remove(filename)


def test_powder_baseline_attributes(powder_dataset):
    """Test that the attributes related to baseline have correct defaults and are
    set to the correct values after computation"""
    assert powder_dataset.first_stage is ""
    assert powder_dataset.wavelet is ""
    assert powder_dataset.level == 0
    assert powder_dataset.niter == 0

    powder_dataset.compute_baseline(
        first_stage="sym6", wavelet="qshift3", level=1, mode="periodic"
    )

    assert powder_dataset.first_stage == "sym6"
    assert powder_dataset.wavelet == "qshift3"
    assert powder_dataset.level == 1


def test_powder_calq(powder_dataset):
    """ Test scattering vector calibration """
    crystal = Crystal.from_database("vo2-m1")
    powder_dataset.powder_calq(crystal, (10, 100), [(1, 0, 0), (2, 0, 0)])

    # Check that shapes match
    assert powder_dataset.powder_eq().shape == powder_dataset.scattering_vector.shape
    # Check that scattering_vector is strictly increasing
    assert np.all(np.greater(np.diff(powder_dataset.scattering_vector), 0))


def test_powder_baseline_limits(powder_dataset):
    """ Test that the baseline is never less than 0, and the baseline-subtracted data is never negative. """

    powder_dataset.compute_baseline(
        first_stage="sym6", wavelet="qshift3", level=1, mode="periodic"
    )

    # Test that the baseline is always positive
    baseline = powder_dataset.powder_baseline(None)
    assert np.all(np.greater_equal(baseline, 0))

    data_bgr = powder_dataset.powder_data(None, bgr=True)
    assert np.all(np.greater_equal(data_bgr, 0))


def test_powder_data_retrieval(powder_dataset):
    """ Test the size of the output from PowderDiffractionDataset.powder_data """
    full_shape = (len(powder_dataset.time_points), powder_dataset.px_radius.size)

    full_data = powder_dataset.powder_data(timedelay=None)
    assert full_data.shape == full_shape

    time_slice = powder_dataset.powder_data(timedelay=powder_dataset.time_points[0])
    assert time_slice.shape == powder_dataset.px_radius.shape


def test_recomputing_angular_average(powder_dataset):
    """Test that recomputing the angular average multiple times will work. This also
    tests resizing all powder data multiple times."""
    powder_dataset.compute_angular_averages(center=(34, 56))
    powder_dataset.compute_baseline(first_stage="sym6", wavelet="qshift1")
    powder_dataset.compute_angular_averages(center=(45, 45), normalized=False)
    powder_dataset.compute_baseline(first_stage="sym5", wavelet="qshift2")
    powder_dataset.compute_angular_averages(center=(34, 56), angular_bounds=(15.3, 187))
    powder_dataset.compute_baseline(first_stage="sym6", wavelet="qshift1")


def test_powder_eq(powder_dataset):
    """ Test PowderDiffractionDataset.powder_eq() """
    eq = powder_dataset.powder_eq()
    assert eq.shape == powder_dataset.px_radius.shape

    powder_dataset.compute_baseline(
        first_stage="sym6", wavelet="qshift3", mode="constant"
    )
    eq = powder_dataset.powder_eq(bgr=True)
    assert eq.shape == powder_dataset.px_radius.shape

    powder_dataset.shift_time_zero(1 + abs(min(powder_dataset.time_points)))
    eq = powder_dataset.powder_eq()
    assert eq.shape == powder_dataset.px_radius.shape
    assert np.allclose(eq, np.zeros_like(eq))



# -*- coding: utf-8 -*-
"""
Update checks management
========================
"""
import json
from urllib.error import URLError
from urllib.request import urlopen

from PyQt5 import QtCore
from packaging.version import Version

from .. import __version__


def update_available():
    """
    Checks whether the currently-installed iris-ued is outdated.

    Returns
    -------
    is_outdated : bool
        Whether or not a new version is available
    latest : str
        Latest available version, currently installed or not.

    Raises
    ------
    ConnectionError
        if connection to PyPI could not be made.
    """
    url = "https://pypi.org/pypi/iris-ued/json"

    try:
        response = urlopen(url).read().decode("utf-8")
    except URLError:
        raise ConnectionError("No connection available.")

    latest_version = Version(json.loads(response)["info"]["version"])

    is_outdated = latest_version > Version(__version__)
    return is_outdated, str(latest_version)


class UpdateChecker(QtCore.QThread):
    """
    Worker that checks for iris-ued updates in a separate thread. Using a separate QThread
    prevents long start times.
    """

    update_available_signal = QtCore.pyqtSignal(bool)
    update_status_signal = QtCore.pyqtSignal(str)

    def __init__(self, *args, **kwargs):
        QtCore.QThread.__init__(self)

    def run(self):
        try:
            outdated, latest = update_available()
            if outdated:
                msg = f"An update is available: latest version is {latest}, and you are currently running version {__version__}."
            else:
                msg = f"You are running the latest version, {__version__}."
        except ConnectionError:
            outdated = False
            msg = "Could not determine if an update is available. No connections available."

        self.update_available_signal.emit(outdated)
        self.update_status_signal.emit(msg)



# -*- coding: utf-8 -*-
"""
Main GUI for iris
"""

from os.path import dirname, join
from pathlib import Path

import pyqtgraph as pg
from PyQt5 import QtCore, QtGui, QtWidgets
from skued import diffread

import qdarkstyle

# Get all proper subclasses of AbstractRawDataset
# to build a loading menu
from .. import AbstractRawDataset, __author__, __license__, __version__
from ..dataset import SWMR_AVAILABLE
from ..plugins import PLUGIN_DIR, install_plugin
from .angular_average_dialog import AngularAverageDialog
from .calibrate_q_dialog import QCalibratorDialog
from .control_bar import ControlBar
from .controller import ErrorAware, IrisController
from .data_viewer import ProcessedDataViewer
from .metadata_edit_dialog import MetadataEditDialog
from .powder_viewer import PowderViewer
from .processing_dialog import ProcessingDialog
from .qbusyindicator import QBusyIndicator
from .qlogger import LOG_DIRECTORY
from .symmetrize_dialog import SymmetrizeDialog
from .update import UpdateChecker

IMAGE_FOLDER = join(dirname(__file__), "images")

INSTALL_PLUGIN_HELP = """You will be prompted to select a plug-in file. This file will be COPIED into:

{dir}

The plug-in is immediately available. The plug-in will remain installed as long as it can be found in the above directory"""


class Iris(QtWidgets.QMainWindow, metaclass=ErrorAware):

    dataset_path_signal = QtCore.pyqtSignal(str)
    raw_dataset_path_signal = QtCore.pyqtSignal(str, object)  # path and class
    single_picture_path_signal = QtCore.pyqtSignal(str)
    error_message_signal = QtCore.pyqtSignal(str)

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.peak_dynamics_window = None

        self._controller_thread = QtCore.QThread(parent=self)
        self.controller = IrisController()  # No parent so that we can moveToThread
        self.controller.moveToThread(self._controller_thread)
        self._controller_thread.start()

        self.dataset_path_signal.connect(self.controller.load_dataset)
        self.raw_dataset_path_signal.connect(self.controller.load_raw_dataset)

        self.controls = ControlBar(parent=self)
        self.controls.raw_data_request.connect(self.controller.display_raw_data)
        self.controls.averaged_data_request.connect(
            self.controller.display_averaged_data
        )
        self.controls.baseline_computation_parameters.connect(
            self.controller.compute_baseline
        )
        self.controls.notes_updated.connect(self.controller.set_dataset_notes)
        self.controls.time_zero_shift.connect(self.controller.set_time_zero_shift)
        self.controls.setSizePolicy(
            QtWidgets.QSizePolicy.Maximum, QtWidgets.QSizePolicy.MinimumExpanding
        )

        self.controller.raw_dataset_loaded_signal.connect(
            lambda b: self.viewer_stack.setCurrentWidget(self.raw_data_viewer)
        )
        self.controller.raw_dataset_loaded_signal.connect(
            self.controls.raw_dataset_controls.setVisible
        )

        self.controller.processed_dataset_loaded_signal.connect(
            lambda b: self.viewer_stack.setCurrentWidget(self.processed_viewer)
        )
        self.controller.processed_dataset_loaded_signal.connect(
            self.controls.diffraction_dataset_controls.setVisible
        )
        self.controller.processed_dataset_loaded_signal.connect(
            self.controls.metadata_and_notes_widget.setVisible
        )

        self.controller.powder_dataset_loaded_signal.connect(
            lambda b: self.viewer_stack.setCurrentWidget(self.powder_viewer)
        )
        self.controller.powder_dataset_loaded_signal.connect(
            self.controls.powder_diffraction_dataset_controls.setVisible
        )

        self.controller.raw_dataset_metadata.connect(
            self.controls.update_raw_dataset_metadata
        )
        self.controller.dataset_metadata.connect(self.controls.update_dataset_metadata)

        # Progress bar --------------------------------------------------------
        self.progress_bar = QtWidgets.QProgressBar(self)
        self.progress_bar.setSizePolicy(
            QtWidgets.QSizePolicy.MinimumExpanding, QtWidgets.QSizePolicy.Maximum
        )
        self.progress_bar.setVisible(False)
        self.controller.processing_data_signal.connect(self.progress_bar.setVisible)

        self.controller.processing_progress_signal.connect(self.progress_bar.setValue)
        self.controller.powder_promotion_progress.connect(self.progress_bar.setValue)
        self.controller.angular_average_progress.connect(self.progress_bar.setValue)

        # Status bar ----------------------------------------------------------
        # Operation in progress widget
        # Including custom 'busy' indicator
        status_bar = QtWidgets.QStatusBar(parent=self)
        self.controller.status_message_signal.connect(
            lambda msg: status_bar.showMessage(msg, 20e3)
        )
        self.setStatusBar(status_bar)

        # Progress bar added to status bar ------------------------------------
        status_bar.addPermanentWidget(self.progress_bar)

        # Busy indicator ------------------------------------------------------
        busy_indicator = QBusyIndicator(parent=self)
        self.controller.operation_in_progress.connect(busy_indicator.toggle_animation)
        status_bar.addPermanentWidget(busy_indicator, 2)

        # Viewers -------------------------------------------------------------
        self.raw_data_viewer = pg.ImageView(parent=self, name="Raw data")
        self.raw_data_viewer.resize(self.raw_data_viewer.maximumSize())
        self.controller.raw_data_signal.connect(
            lambda obj: self.raw_data_viewer.clear()
            if obj is None
            else self.raw_data_viewer.setImage(obj, autoLevels=False)
        )

        self.processed_viewer = ProcessedDataViewer(parent=self)
        self.processed_viewer.timeseries_rect_signal.connect(
            self.controller.time_series
        )
        self.controller.averaged_data_signal.connect(self.processed_viewer.display)
        self.controller.time_series_signal.connect(
            self.processed_viewer.display_peak_dynamics
        )
        self.controller.bragg_peaks_signal.connect(
            self.processed_viewer.display_bragg_peaks
        )

        self.powder_viewer = PowderViewer(parent=self)
        self.powder_viewer.peak_dynamics_roi_signal.connect(
            self.controller.powder_time_series
        )
        self.controller.powder_data_signal.connect(
            self.powder_viewer.display_powder_data
        )
        self.controller.powder_time_series_signal.connect(
            self.powder_viewer.display_peak_dynamics
        )
        # UI components
        self.controller.error_message_signal.connect(self.show_error_message)
        self.error_message_signal.connect(self.show_error_message)

        self.file_dialog = QtWidgets.QFileDialog(parent=self)
        self.menu_bar = self.menuBar()

        self.viewer_stack = QtWidgets.QTabWidget()
        self.viewer_stack.addTab(self.raw_data_viewer, "View raw dataset")
        self.viewer_stack.addTab(self.processed_viewer, "View processed dataset")
        self.viewer_stack.addTab(self.powder_viewer, "View azimuthal averages")

        self.controller.raw_dataset_loaded_signal.connect(
            lambda toggle: self.viewer_stack.setTabEnabled(0, toggle)
        )
        self.controller.processed_dataset_loaded_signal.connect(
            lambda toggle: self.viewer_stack.setTabEnabled(1, toggle)
        )
        self.controller.powder_dataset_loaded_signal.connect(
            lambda toggle: self.viewer_stack.setTabEnabled(2, toggle)
        )

        ###################
        # Actions

        self.file_menu = self.menu_bar.addMenu("&File")
        self.file_menu.setToolTipsVisible(True)
        self.load_raw_submenu = self.file_menu.addMenu("Load raw dataset...")

        # Create the menu based on the currently-available plugins
        # This can be called multiple times (e.g. when new plugins are installed)
        self.create_load_raw_menu()

        self.load_dataset_action = QtWidgets.QAction(
            QtGui.QIcon(join(IMAGE_FOLDER, "locator.png")), "& Load dataset", self
        )
        self.load_dataset_action.triggered.connect(self.load_dataset)

        self.load_single_picture_action = QtWidgets.QAction(
            QtGui.QIcon(join(IMAGE_FOLDER, "locator.png")),
            "& Load diffraction picture",
            self,
        )
        self.load_single_picture_action.triggered.connect(self.load_single_picture)

        self.close_raw_dataset_action = QtWidgets.QAction(
            QtGui.QIcon(join(IMAGE_FOLDER, "locator.png")), "& Close raw dataset", self
        )
        self.close_raw_dataset_action.triggered.connect(
            self.controller.close_raw_dataset
        )
        self.close_raw_dataset_action.setEnabled(False)
        self.controller.raw_dataset_loaded_signal.connect(
            self.close_raw_dataset_action.setEnabled
        )

        self.close_dataset_action = QtWidgets.QAction(
            QtGui.QIcon(join(IMAGE_FOLDER, "locator.png")), "& Close dataset", self
        )
        self.close_dataset_action.triggered.connect(self.controller.close_dataset)
        self.close_dataset_action.setEnabled(False)
        self.controller.powder_dataset_loaded_signal.connect(
            self.close_dataset_action.setEnabled
        )
        self.controller.processed_dataset_loaded_signal.connect(
            self.close_dataset_action.setEnabled
        )

        self.file_menu.addAction(self.load_dataset_action)
        self.file_menu.addAction(self.load_single_picture_action)
        self.file_menu.addSeparator()
        self.file_menu.addAction(self.close_raw_dataset_action)
        self.file_menu.addAction(self.close_dataset_action)

        ###################
        # Plug-in Actions
        self.install_plugin_action = QtWidgets.QAction(
            QtGui.QIcon(join(IMAGE_FOLDER, "eye.png")), "& Install plug-in", self
        )
        self.install_plugin_action.setToolTip(
            "Copy a plug-in file into the internal storage. The new plug-in will be available immediately."
        )
        self.install_plugin_action.triggered.connect(self.install_plugin)

        self.open_plugin_directory_action = QtWidgets.QAction(
            QtGui.QIcon(join(IMAGE_FOLDER, "eye.png")), "& Open plug-in directory", self
        )
        self.open_plugin_directory_action.triggered.connect(
            lambda: QtGui.QDesktopServices.openUrl(
                QtCore.QUrl("file:///" + str(PLUGIN_DIR), QtCore.QUrl.TolerantMode)
            )
        )

        self.howto_write_plugin_action = QtWidgets.QAction(
            QtGui.QIcon(join(IMAGE_FOLDER, "revert.png")), "& Writing a plug-in", self
        )
        self.howto_write_plugin_action.triggered.connect(
            lambda: QtGui.QDesktopServices.openUrl(
                QtCore.QUrl("http://iris-ued.readthedocs.io/en/master/plugins.html")
            )
        )

        self.plugin_menu = self.menu_bar.addMenu("Plug-ins")
        self.plugin_menu.addAction(self.install_plugin_action)
        self.plugin_menu.addAction(self.open_plugin_directory_action)
        self.plugin_menu.addSeparator()
        self.plugin_menu.addAction(self.howto_write_plugin_action)

        ###################
        # Operations on Diffraction Datasets
        self.processing_action = QtWidgets.QAction(
            QtGui.QIcon(join(IMAGE_FOLDER, "analysis.png")), "& Process raw data", self
        )
        self.processing_action.triggered.connect(self.launch_processsing_dialog)
        self.controller.raw_dataset_loaded_signal.connect(
            self.processing_action.setEnabled
        )

        self.symmetrize_action = QtWidgets.QAction(
            QtGui.QIcon(join(IMAGE_FOLDER, "analysis.png")), "& Symmetrize data", self
        )
        self.symmetrize_action.triggered.connect(self.launch_symmetrize_dialog)
        self.controller.processed_dataset_loaded_signal.connect(
            self.symmetrize_action.setEnabled
        )

        self.calculate_azimuthal_averages_action = QtWidgets.QAction(
            QtGui.QIcon(join(IMAGE_FOLDER, "analysis.png")),
            "& Calculate azimuthal averages",
            self,
        )
        self.calculate_azimuthal_averages_action.triggered.connect(
            self.launch_calculate_azimuthal_averages_dialog
        )
        self.controller.processed_dataset_loaded_signal.connect(
            self.calculate_azimuthal_averages_action.setEnabled
        )

        self.update_metadata_action = QtWidgets.QAction(
            QtGui.QIcon(join(IMAGE_FOLDER, "save.png")),
            "& Update dataset metadata",
            self,
        )
        self.update_metadata_action.triggered.connect(self.launch_metadata_edit_dialog)
        self.controller.processed_dataset_loaded_signal.connect(
            self.update_metadata_action.setEnabled
        )

        self.calibrate_scattvector_action = QtWidgets.QAction(
            QtGui.QIcon(join(IMAGE_FOLDER, "analysis.png")),
            "& Calibrate scattering vector",
            self,
        )
        self.calibrate_scattvector_action.triggered.connect(self.launch_calq_dialog)
        self.controller.powder_dataset_loaded_signal.connect(
            self.calibrate_scattvector_action.setEnabled
        )

        self.diffraction_dataset_menu = self.menu_bar.addMenu("&Dataset")
        self.diffraction_dataset_menu.addAction(self.processing_action)
        self.diffraction_dataset_menu.addSeparator()
        self.diffraction_dataset_menu.addAction(self.update_metadata_action)
        self.diffraction_dataset_menu.addAction(self.symmetrize_action)
        self.diffraction_dataset_menu.addAction(
            self.calculate_azimuthal_averages_action
        )
        self.diffraction_dataset_menu.addSeparator()
        self.diffraction_dataset_menu.addAction(self.calibrate_scattvector_action)

        ###################
        # Display options
        self.toggle_controls_visibility_action = QtWidgets.QAction(
            "& Show/hide controls", self
        )
        self.toggle_controls_visibility_action.setCheckable(True)
        self.toggle_controls_visibility_action.setChecked(True)
        self.toggle_controls_visibility_action.toggled.connect(self.controls.setVisible)
        self.show_diff_peak_dynamics_action = QtWidgets.QAction(
            "& Show/hide peak dynamics", self
        )
        self.show_diff_peak_dynamics_action.setCheckable(True)
        self.show_diff_peak_dynamics_action.toggled.connect(
            self.processed_viewer.toggle_peak_dynamics
        )
        self.controller.processed_dataset_loaded_signal.connect(
            self.show_diff_peak_dynamics_action.setEnabled
        )

        self.show_diff_peak_dynamics_bounds_action = QtWidgets.QAction(
            "& Show/hide peak dynamics bounds", self
        )
        self.show_diff_peak_dynamics_bounds_action.setCheckable(True)
        self.show_diff_peak_dynamics_bounds_action.toggled.connect(
            self.processed_viewer.toggle_roi_bounds_text
        )
        self.controller.processed_dataset_loaded_signal.connect(
            self.show_diff_peak_dynamics_bounds_action.setEnabled
        )

        self.show_diff_relative_action = QtWidgets.QAction(
            "& Toggle relative dynamics", self
        )
        self.show_diff_relative_action.setCheckable(True)
        self.controller.relative_averaged_enable_signal.connect(
            self.show_diff_relative_action.setChecked
        )
        self.show_diff_relative_action.toggled.connect(
            self.controller.enable_averaged_relative
        )
        self.controller.processed_dataset_loaded_signal.connect(
            self.show_diff_relative_action.setEnabled
        )

        self.show_bragg_peaks_action = QtWidgets.QAction("& Show Bragg peaks", self)
        self.controller.processed_dataset_loaded_signal.connect(
            self.show_bragg_peaks_action.setEnabled
        )
        self.show_bragg_peaks_action.triggered.connect(self.controller.find_bragg_peaks)

        self.show_powder_relative_action = QtWidgets.QAction(
            "& Toggle relative dynamics", self
        )
        self.show_powder_relative_action.setCheckable(True)
        self.controller.relative_powder_enable_signal.connect(
            self.show_powder_relative_action.setChecked
        )
        self.show_powder_relative_action.toggled.connect(
            self.controller.enable_powder_relative
        )
        self.controller.powder_dataset_loaded_signal.connect(
            self.show_powder_relative_action.setEnabled
        )

        self.toggle_powder_background_action = QtWidgets.QAction(
            "& Remove baseline", self
        )
        self.toggle_powder_background_action.setCheckable(True)
        self.controller.powder_bgr_enable_signal.connect(
            self.toggle_powder_background_action.setChecked
        )
        self.toggle_powder_background_action.toggled.connect(
            self.controller.powder_background_subtracted
        )
        self.controller.powder_dataset_loaded_signal.connect(
            self.toggle_powder_background_action.setEnabled
        )

        self.display_options_menu = self.menu_bar.addMenu("&Display")
        self.display_options_menu.addAction(self.toggle_controls_visibility_action)
        self.display_options_menu.addSeparator()
        self.diffraction_dataset_display_options_menu = (
            self.display_options_menu.addMenu("& Diffraction display options")
        )
        self.controller.processed_dataset_loaded_signal.connect(
            self.diffraction_dataset_display_options_menu.setEnabled
        )

        self.powder_dataset_display_options_menu = self.display_options_menu.addMenu(
            "& Powder display options"
        )
        self.controller.powder_dataset_loaded_signal.connect(
            self.powder_dataset_display_options_menu.setEnabled
        )

        # Assemble submenus
        self.diffraction_dataset_display_options_menu.addAction(
            self.show_diff_peak_dynamics_action
        )
        self.diffraction_dataset_display_options_menu.addAction(
            self.show_diff_peak_dynamics_bounds_action
        )
        self.diffraction_dataset_display_options_menu.addAction(
            self.show_diff_relative_action
        )
        self.diffraction_dataset_display_options_menu.addAction(
            self.show_bragg_peaks_action
        )

        self.powder_dataset_display_options_menu.addAction(
            self.show_powder_relative_action
        )
        self.powder_dataset_display_options_menu.addAction(
            self.toggle_powder_background_action
        )

        ###################
        # Helps and misc operations
        self.about_action = QtWidgets.QAction("& About", self)
        self.about_action.triggered.connect(self.show_about)

        self.launch_documentation_action = QtWidgets.QAction(
            QtGui.QIcon(join(IMAGE_FOLDER, "revert.png")),
            "& Open online documentation",
            self,
        )
        self.launch_documentation_action.triggered.connect(
            lambda: QtGui.QDesktopServices.openUrl(
                QtCore.QUrl("http://iris-ued.readthedocs.io/en/master/")
            )
        )

        self.goto_repository_action = QtWidgets.QAction(
            QtGui.QIcon(join(IMAGE_FOLDER, "revert.png")),
            "& Go to GitHub Repository",
            self,
        )
        self.goto_repository_action.triggered.connect(
            lambda: QtGui.QDesktopServices.openUrl(
                QtCore.QUrl("https://github.com/LaurentRDC/iris-ued/")
            )
        )

        self.report_issue_action = QtWidgets.QAction(
            QtGui.QIcon(join(IMAGE_FOLDER, "revert.png")), "& Report issue", self
        )
        self.report_issue_action.triggered.connect(
            lambda: QtGui.QDesktopServices.openUrl(
                QtCore.QUrl("https://github.com/LaurentRDC/iris-ued/issues/new")
            )
        )

        self.open_log_directory_action = QtWidgets.QAction("& Open log directory", self)
        self.open_log_directory_action.triggered.connect(
            lambda: QtGui.QDesktopServices.openUrl(
                QtCore.QUrl("file:///" + str(LOG_DIRECTORY))
            )
        )

        # When iris starts, no knowledge if update is available
        self.update_action = QtWidgets.QAction("& Checking for updates...", self)
        self.update_action.triggered.connect(
            lambda: QtGui.QDesktopServices.openUrl(
                QtCore.QUrl("https://github.com/LaurentRDC/iris-ued/releases/latest")
            )
        )
        self.update_action.setEnabled(
            False
        )  # Signal to update this will be done in background

        self.help_menu = self.menu_bar.addMenu("&Help")
        self.help_menu.addAction(self.about_action)
        self.help_menu.addSeparator()
        self.help_menu.addAction(self.launch_documentation_action)
        self.help_menu.addAction(self.goto_repository_action)
        self.help_menu.addAction(self.report_issue_action)
        self.help_menu.addSeparator()
        self.help_menu.addAction(self.open_log_directory_action)
        self.help_menu.addAction(self.update_action)

        ###################
        # Layout
        self.layout = QtWidgets.QHBoxLayout()
        self.layout.addWidget(self.viewer_stack)
        self.layout.addWidget(self.controls)

        self.central_widget = QtWidgets.QWidget()
        self.central_widget.setLayout(self.layout)
        self.setCentralWidget(self.central_widget)

        # Initialization
        self.controller.raw_dataset_loaded_signal.emit(False)
        self.controller.powder_dataset_loaded_signal.emit(False)
        self.controller.processed_dataset_loaded_signal.emit(False)
        self.controller.processing_progress_signal.emit(0)
        self.controller.powder_promotion_progress.emit(0)
        self.controller.angular_average_progress.emit(0)

        self.viewer_stack.setCurrentWidget(self.raw_data_viewer)
        self.viewer_stack.resize(self.viewer_stack.maximumSize())

        # Window settings ------------------------------------------------------
        # Set un-maximizzed geometry to be 75% of availiable screen space
        available = QtWidgets.QApplication.desktop().availableGeometry()
        available.setSize(0.75 * available.size())
        self.setGeometry(available)

        self.setWindowIcon(QtGui.QIcon(join(IMAGE_FOLDER, "eye.png")))
        self.setWindowTitle("Iris - UED data exploration")
        self.center_window()
        self.showMaximized()

        # At the end, we start the check for an update
        # This is done in a separate thread to prevent slow startups
        # Note: the update status signal is passed to the controller
        # so it can be logged as well.
        self.update_checker = UpdateChecker(parent=self)
        self.update_checker.update_status_signal.connect(
            self.controller.status_message_signal
        )
        self.update_checker.update_available_signal.connect(self.update_available)
        self.update_checker.start()

    def create_load_raw_menu(self):
        # Dynamically add an option for each implementation of AbstractRawDataset
        # Note : because of dynamical nature of these bindings,
        # it must be done in a separate method (_create_load_raw)
        # Otherwise, it doesn't work correctly
        self.load_raw_submenu.clear()

        if len(AbstractRawDataset.implementations) == 0:
            act = QtWidgets.QAction("& (no plug-ins installed)", parent=self)
            act.setEnabled(False)
            self.load_raw_submenu.addAction(act)
        else:
            for rawdatacls in sorted(
                AbstractRawDataset.implementations, key=lambda c: c.display_name
            ):
                self._create_load_raw(rawdatacls, self.load_raw_submenu)

    def _create_load_raw(self, cls, submenu):
        submenu.addAction(
            QtGui.QIcon(join(IMAGE_FOLDER, "locator.png")),
            f"&{cls.display_name}",
            lambda: self.load_raw_dataset(cls),
        )

    def closeEvent(self, event):
        self._controller_thread.quit()
        super().closeEvent(event)

    @QtCore.pyqtSlot(str)
    def show_error_message(self, msg):
        self.error_dialog = QtWidgets.QErrorMessage(parent=self)
        self.error_dialog.showMessage(msg)

    @QtCore.pyqtSlot()
    def launch_processsing_dialog(self):
        processing_dialog = ProcessingDialog(
            parent=self, raw=self.controller.raw_dataset
        )
        processing_dialog.resize(0.75 * self.size())
        processing_dialog.processing_parameters_signal.connect(
            self.controller.process_raw_dataset
        )
        processing_dialog.exec_()
        processing_dialog.processing_parameters_signal.disconnect(
            self.controller.process_raw_dataset
        )

    @QtCore.pyqtSlot()
    def launch_symmetrize_dialog(self):
        symmetrize_dialog = SymmetrizeDialog(
            parent=self,
            image=self.controller.dataset.diff_data(
                timedelay=self.controller.dataset.time_points[0]
            ),
            mask=self.controller.dataset.valid_mask,
            center=self.controller.dataset.center,
        )
        symmetrize_dialog.resize(0.75 * self.size())
        symmetrize_dialog.symmetrize_parameters_signal.connect(
            self.controller.symmetrize
        )
        symmetrize_dialog.exec_()
        symmetrize_dialog.symmetrize_parameters_signal.disconnect(
            self.controller.symmetrize
        )

    @QtCore.pyqtSlot()
    def launch_metadata_edit_dialog(self):
        metadata_dialog = MetadataEditDialog(
            parent=self, config=self.controller.dataset.metadata
        )
        metadata_dialog.updated_metadata_signal.connect(self.controller.update_metadata)
        metadata_dialog.exec_()
        metadata_dialog.updated_metadata_signal.disconnect(
            self.controller.update_metadata
        )

    @QtCore.pyqtSlot()
    def launch_calculate_azimuthal_averages_dialog(self):
        promote_dialog = AngularAverageDialog(
            self.controller.dataset.diff_data(self.controller.dataset.time_points[0]),
            mask=self.controller.dataset.valid_mask,
            center=self.controller.dataset.center,
            parent=self,
        )
        promote_dialog.resize(0.75 * self.size())
        promote_dialog.angular_average_signal.connect(
            self.controller.calculate_azimuthal_averages
        )
        promote_dialog.exec_()
        promote_dialog.angular_average_signal.disconnect(
            self.controller.calculate_azimuthal_averages
        )

    @QtCore.pyqtSlot()
    def launch_recompute_angular_average_dialog(self):
        dialog = AngularAverageDialog(
            self.controller.dataset.diff_data(self.controller.dataset.time_points[0]),
            mask=self.controller.dataset.valid_mask,
            center=self.controller.dataset.center,
            parent=self,
        )
        dialog.resize(0.75 * self.size())
        dialog.angular_average_signal.connect(self.controller.recompute_angular_average)
        dialog.exec_()
        dialog.angular_average_signal.disconnect(
            self.controller.recompute_angular_average
        )

    @QtCore.pyqtSlot()
    def launch_calq_dialog(self):
        I = self.controller.dataset.powder_eq(bgr=True)
        dialog = QCalibratorDialog(I, parent=self)
        dialog.resize(0.75 * self.size())
        dialog.calibration_parameters.connect(self.controller.powder_calq)
        dialog.exec_()
        dialog.calibration_parameters.disconnect(self.controller.powder_calq)

    @QtCore.pyqtSlot(bool)
    def update_available(self, available):
        """ Handle UI in case an update is available or not. """
        if available:
            self.update_action.setEnabled(True)
            self.update_action.setText("An update is available!")
        else:
            self.update_action.setEnabled(False)
            self.update_action.setText("No updates available.")

    @QtCore.pyqtSlot(object)
    def load_raw_dataset(self, cls):
        path = self.file_dialog.getExistingDirectory(
            parent=self, caption="Load raw dataset"
        )
        if not path:
            return
        self.raw_dataset_path_signal.emit(path, cls)

    @QtCore.pyqtSlot()
    def load_dataset(self):
        path = self.file_dialog.getOpenFileName(
            parent=self, caption="Load dataset", filter="*.hdf5"
        )[0]
        if not path:
            return
        self.dataset_path_signal.emit(path)

    @QtCore.pyqtSlot()
    def load_single_picture(self):
        path = self.file_dialog.getOpenFileName(
            parent=self,
            caption="Load diffraction picture",
            filter="Images (*.tif *.tiff *.mib *.dm3 *.dm4)",
        )[0]
        if not path:
            return

        self.image_viewer = pg.ImageView(parent=self)
        self.image_viewer.setImage(diffread(path))

        layout = QtWidgets.QVBoxLayout()
        fname_label = QtWidgets.QLabel("Filename: " + path, parent=self)
        fname_label.setAlignment(QtCore.Qt.AlignCenter)
        layout.addWidget(fname_label)
        layout.addWidget(self.image_viewer)

        dialog = QtWidgets.QDialog()
        dialog.setLayout(layout)
        dialog.resize(0.75 * self.size())
        return dialog.exec_()

    @QtCore.pyqtSlot()
    def center_window(self):
        qr = self.frameGeometry()
        cp = QtWidgets.QDesktopWidget().availableGeometry().center()
        qr.moveCenter(cp)
        self.move(qr.topLeft())

    @QtCore.pyqtSlot()
    def install_plugin(self):
        """ Load plug-in. """
        explanation = INSTALL_PLUGIN_HELP.format(dir=PLUGIN_DIR)

        QtWidgets.QMessageBox.information(self, "Loading a plug-in", explanation)

        path = self.file_dialog.getOpenFileName(
            parent=self, caption="Load plug-in file", filter="Python source (*.py)"
        )[0]
        if not path:
            return

        install_plugin(path)
        self.create_load_raw_menu()

    @QtCore.pyqtSlot()
    def show_about(self):
        """ Show the About information """

        return QtWidgets.QMessageBox.about(self, "About Iris", make_about_string())


def make_about_string():
    import h5py
    import sys
    import skued
    import pyqtgraph
    import npstreams

    python_version = "{}.{}.{}".format(*sys.version_info[:3])

    return f"""<h2>About Iris</h2>
    Iris is both a Python library and a GUI program for the exploration of ultrafast electron diffraction data. <br>
    <b>License</b>: {__license__}                    <br>
    <b>Author</b>: {__author__}                      <br>
    <b>Install location</b>: {Path(__file__).parent} <br> 
    <b>SWMR available</b>: {SWMR_AVAILABLE}          <br>

    <h3>Versions</h3>
    <table style="width:100%">
        <tr>
            <td style="padding-right:75">Python</td>
            <td align="right">{python_version}</td>
        </tr>
        <tr>
            <td style="padding-right:75">Iris</td>
            <td align="right">{__version__}</td>
        </tr>
        <tr>
            <td style="padding-right:75">Qt</td>
            <td align="right">{QtCore.qVersion()}</td>
        </tr>
        <tr>
            <td style="padding-right:75">qdarkstyle</td>
            <td align="right">{qdarkstyle.__version__}</td>
        </tr>
        <tr>
            <td style="padding-right:75">PyQtGraph</td>
            <td align="right">{pyqtgraph.__version__}</td>
        </tr>
        <tr>
            <td style="padding-right:75">scikit-ued</td>
            <td align="right">{skued.__version__}</td>
        </tr>
        <tr>
            <td style="padding-right:75">npstreams</td>
            <td align="right">{npstreams.__version__}</td>
        </tr>
        <tr>
            <td style="padding-right:75">HDF5</td>
            <td align="right">{h5py.version.hdf5_version}</td>
        </tr>
        <tr>
            <td style="padding-right:75">h5py</td>
            <td align="right">{h5py.version.version}</td>
        </tr>
    </table>

    <h3>Installed plug-ins</h3>
    {'<br>'.join(cls.__name__ for cls in sorted(AbstractRawDataset.implementations,  key = lambda cls: cls.__name__))} """



# -*- coding: utf-8 -*-
"""
Test plug-in
============
"""

from iris import AbstractRawDataset, ExperimentalParameter
import numpy as np


class TestRawDatasetPlugin(AbstractRawDataset):

    display_name = "Minimal raw dataset v1"

    temperature = ExperimentalParameter("temperature", ptype=float, default=500)
    is_useful = ExperimentalParameter("is_useful", ptype=bool, default=True)

    # We don't want pytest to collect this class as a test
    # https://stackoverflow.com/a/63430765
    __test__ = False

    def __init__(self, source=None, metadata=dict()):
        # Metadata can be filled as a dictionary before
        # initialization. # Attributes which are not
        # ExperimentalParameters are ignored.
        metadata.update(
            {"temperature": 100, "exposure": 1, "this_will_be_ignored": True}
        )
        super().__init__(source, metadata)

        self.time_points = tuple(range(0, 3))
        self.resolution = (64, 64)
        self.scans = tuple(range(1, 3))
        self.notes = "This is a test"
        self.is_useful = False

    def raw_data(self, *args, **kwargs):
        return np.random.random(size=self.resolution)



# -*- coding: utf-8 -*-
"""
Diffraction dataset types
"""
from sys import platform
from collections import OrderedDict
from collections.abc import Callable
from functools import lru_cache, partial, wraps
from math import sqrt
from warnings import warn

import h5py
import numpy as np
from scipy.ndimage import gaussian_filter

import npstreams as ns
from skued import (
    __version__,
    nfold,
    autocenter,
    ArbitrarySelection,
    Selection,
)

from .meta import HDF5ExperimentalParameter, MetaHDF5Dataset

# Whether or not single-writer multiple-reader (SWMR) mode is available
# See http://docs.h5py.org/en/latest/swmr.html for more information
SWMR_AVAILABLE = h5py.version.hdf5_version_tuple > (1, 10, 0)


class MigrationWarning(UserWarning):
    """ Warning class for warnings involving the migration of datasets to a newer version. """

    pass


class MigrationError(Exception):
    """ Thrown if a particular dataset requires migration. """

    pass


def write_access_needed(f):
    """ Ensure that write access has been granted before using a method. """

    @wraps(f)
    def newf(self, *args, **kwargs):
        if self.mode != "r+":
            raise PermissionError(
                f"The dataset {self.filename} has not been opened with write access."
            )
        return f(self, *args, **kwargs)

    return newf


def update_center(f):
    """ Recompute the dependent quantities following a transformation, i.e. equilibrium pattern. """

    @wraps(f)
    def newf(self, *args, **kwargs):
        r = f(self, *args, **kwargs)
        self._autocenter()
        return r

    return newf


def update_equilibrium_pattern(f):
    """ Recompute the dependent quantities following a transformation, i.e. equilibrium pattern. """

    @wraps(f)
    def newf(self, *args, **kwargs):
        r = f(self, *args, **kwargs)
        self._recompute_diff_eq()  # It is assumed that diff_eq caches the result
        return r

    return newf


class DiffractionDataset(h5py.File, metaclass=MetaHDF5Dataset):
    """
    Abstraction of an HDF5 file to represent diffraction datasets.
    """

    _diffraction_group_name = "/processed"
    _exp_params_group_name = "/"

    # Subclasses can add more experimental parameters like those below
    # The types must be representable by h5py
    center = HDF5ExperimentalParameter("center", tuple, default=(0, 0))
    acquisition_date = HDF5ExperimentalParameter(
        "acquisition_date", str, default=""
    )  # Acquisition date, no specific format
    energy = HDF5ExperimentalParameter("energy", float, default=90)  # keV
    pump_wavelength = HDF5ExperimentalParameter(
        "pump_wavelength", int, default=800
    )  # nanometers
    fluence = HDF5ExperimentalParameter(
        "fluence", float, default=0
    )  # milliJoules / centimeters ^ 2
    time_zero_shift = HDF5ExperimentalParameter(
        "time_zero_shift", float, default=0
    )  # picoseconds
    temperature = HDF5ExperimentalParameter(
        "temperature", float, default=293
    )  # kelvins
    exposure = HDF5ExperimentalParameter("exposure", float, default=1)  # seconds
    scans = HDF5ExperimentalParameter("scans", tuple, default=(1,))
    camera_length = HDF5ExperimentalParameter(
        "camera_length", float, default=0.23
    )  # meters
    pixel_width = HDF5ExperimentalParameter(
        "pixel_width", float, default=14e-6
    )  # meters
    aligned = HDF5ExperimentalParameter("aligned", bool, default=False)
    normalized = HDF5ExperimentalParameter("normalized", bool, default=False)
    notes = HDF5ExperimentalParameter("notes", str, default="")

    def __init__(self, *args, **kwargs):
        # Secret option to skip the checks below
        # This is only useful when building a dataset
        # Don't use it
        skip_checks = kwargs.pop("skip_checks", False)

        super().__init__(*args, **kwargs)

        if not skip_checks:
            self._migration_checks()

    def _migration_checks(self):
        """
        Migration checks should be performed here. As iris has evolved beyond v5.0.0,
        new requirements have emerged. If the file is opened with writing
        permissions, we can migrate silently.
        """
        if self.center == (0, 0):  # default
            if self.mode == "r+":
                self._autocenter()
            else:
                warn(
                    "".join(
                        [
                            f"The center of diffraction for the dataset {self.filename} is missing.",
                            "Open it with writing permissions so it can be calculated.",
                            "This warning will become an error in future versions of iris.",
                        ]
                    ),
                    category=MigrationWarning,
                    stacklevel=2,
                )
        if "equilibrium" not in self.diffraction_group:
            if self.mode == "r+":
                self._recompute_diff_eq()
            else:
                warn(
                    "".join(
                        [
                            f"The equilibrium diffraction pattern for the dataset {self.filename} is missing.",
                            "Open it with writing permissions so it can be calculated.",
                            "This warning will become an error in future versions of iris.",
                        ]
                    ),
                    category=MigrationWarning,
                    stacklevel=2,
                )

    def __repr__(self):
        rep = f"< {type(self).__name__} object with following metadata: "
        for key, val in self.metadata.items():
            rep += f"\n    {key}: {val}"

        return rep + " >"

    @classmethod
    def from_collection(
        cls,
        patterns,
        filename,
        time_points,
        metadata,
        valid_mask=None,
        dtype=None,
        ckwargs=None,
        callback=None,
        **kwargs,
    ):
        """
        Create a DiffractionDataset from a collection of diffraction patterns and metadata.

        Parameters
        ----------
        patterns : iterable of ndarray or ndarray
            Diffraction patterns. These should be in the same order as ``time_points``. Note that
            the iterable can be a generator, in which case it will be consumed.
        filename : str or path-like
            Path to the assembled DiffractionDataset.
        time_points : array_like, shape (N,)
            Time-points of the diffraction patterns, in picoseconds.
        metadata : dict
            Valid keys are contained in ``DiffractionDataset.valid_metadata``.
        valid_mask : ndarray or None, optional
            Boolean array that evaluates to True on valid pixels. This information is useful in
            cases where a beamblock is used.
        dtype : dtype or None, optional
            Patterns will be cast to ``dtype``. If None (default), ``dtype`` will be set to the same
            data-type as the first pattern in ``patterns``.
        ckwargs : dict, optional
            HDF5 compression keyword arguments. Refer to ``h5py``'s documentation for details.
            Default is to use the `lzf` compression pipeline.
        callback : callable or None, optional
            Callable that takes an int between 0 and 99. This can be used for progress update when
            ``patterns`` is a generator and involves large computations.
        kwargs
            Keywords are passed to ``h5py.File`` constructor.
            Default is file-mode 'x', which raises error if file already exists.
            Default libver is 'latest'.

        Returns
        -------
        dataset : DiffractionDataset
        """
        if "mode" not in kwargs:
            kwargs["mode"] = "x"

        if "libver" not in kwargs:
            kwargs["libver"] = "latest"

        # H5py will raise an exception if arrays are not contiguous
        # patterns = map(np.ascontiguousarray, iter(patterns))

        if callback is None:
            callback = lambda _: None

        time_points = np.array(time_points).reshape(-1)

        if ckwargs is None:
            ckwargs = {"compression": "lzf", "shuffle": True, "fletcher32": True}
        ckwargs[
            "chunks"
        ] = True  # For some reason, if no chunking, writing to disk is SLOW

        first, patterns = ns.peek(patterns)
        if dtype is None:
            dtype = first.dtype
        resolution = first.shape

        if valid_mask is None:
            valid_mask = np.ones(first.shape, dtype=bool)

        callback(0)
        with cls(filename, skip_checks=True, **kwargs) as file:

            # Note that keys not associated with an ExperimentalParameter
            # descriptor will not be recorded in the file.
            metadata.pop("time_points", None)
            for key, val in metadata.items():
                if key not in cls.valid_metadata:
                    continue
                setattr(file, key, val)

            # Record time-points as a dataset; then, changes to it will be reflected
            # in other dimension scales
            gp = file.experimental_parameters_group
            times = gp.create_dataset("time_points", data=time_points, dtype=float)
            mask = gp.create_dataset("valid_mask", data=valid_mask, dtype=bool)

            pgp = file.diffraction_group
            dset = pgp.create_dataset(
                name="intensity",
                shape=resolution + (len(time_points),),
                dtype=dtype,
                **ckwargs,
            )

            # Making use of the H5DS dimension scales
            # http://docs.h5py.org/en/latest/high/dims.html
            times.make_scale("time-delay")
            dset.dims[2].attach_scale(times)

            # At each iteration, we flush the changes to file
            # If this is not done, data can be accumulated in memory (>5GB)
            # until this loop is done.
            for index, pattern in enumerate(patterns):
                dset.write_direct(pattern, dest_sel=np.s_[:, :, index])
                file.flush()
                callback(round(100 * index / np.size(time_points)))

            file._autocenter()
            file._recompute_diff_eq()

        callback(100)

        # Now that the file exists, we can switch to read/write mode
        kwargs["mode"] = "r+"
        return cls(filename, **kwargs)

    @classmethod
    def from_raw(
        cls,
        raw,
        filename,
        exclude_scans=None,
        valid_mask=None,
        processes=1,
        callback=None,
        align=True,
        normalize=True,
        ckwargs=None,
        dtype=None,
        **kwargs,
    ):
        """
        Create a DiffractionDataset from a subclass of AbstractRawDataset.

        Parameters
        ----------
        raw : AbstractRawDataset instance
            Raw dataset instance.
        filename : str or path-like
            Path to the assembled DiffractionDataset.
        exclude_scans : iterable of ints or None, optional
            Scans to exclude from the processing. Default is to include all scans.
        valid_mask : ndarray or None, optional
            Boolean array that evaluates to True on valid pixels. This information is useful in
            cases where a beamblock is used.
        processes : int or None, optional
            Number of Processes to spawn for processing. Default is number of available
            CPU cores.
        callback : callable or None, optional
            Callable that takes an int between 0 and 99. This can be used for progress update.
        align : bool, optional
            If True (default), raw images will be aligned on a per-scan basis.
        normalize : bool, optional
            If True, images within a scan are normalized to the same integrated diffracted intensity.
        ckwargs : dict or None, optional
            HDF5 compression keyword arguments. Refer to ``h5py``'s documentation for details.
        dtype : dtype or None, optional
            Patterns will be cast to ``dtype``. If None (default), ``dtype`` will be set to the same
            data-type as the first pattern in ``patterns``.
        kwargs
            Keywords are passed to ``h5py.File`` constructor.
            Default is file-mode 'x', which raises error if file already exists.

        Returns
        -------
        dataset : DiffractionDataset

        See Also
        --------
        open_raw : open raw datasets by guessing the appropriate format based on available plug-ins.

        Raises
        ------
        IOError
            If the filename is already associated with a file.
        """
        if callback is None:
            callback = lambda _: None

        if exclude_scans is None:
            exclude_scans = set([])

        if valid_mask is None:
            valid_mask = np.ones(shape=raw.resolution, dtype=bool)

        metadata = raw.metadata.copy()
        metadata["scans"] = tuple(set(raw.scans) - set(exclude_scans))
        metadata["aligned"] = align
        metadata["normalized"] = normalize

        # Assemble the metadata
        kwargs.update(
            {
                "ckwargs": ckwargs,
                "valid_mask": valid_mask,
                "metadata": metadata,
                "time_points": raw.time_points,
                "dtype": dtype,
                "callback": callback,
                "filename": filename,
            }
        )

        reduced = raw.reduced(
            exclude_scans=exclude_scans,
            align=align,
            normalize=normalize,
            mask=np.logical_not(valid_mask),
            processes=processes,
            dtype=dtype,
        )

        return cls.from_collection(patterns=reduced, **kwargs)

    @write_access_needed
    @update_center
    @update_equilibrium_pattern
    def diff_apply(self, func, callback=None, processes=1):
        """
        Apply a function to each diffraction pattern possibly in parallel. The diffraction patterns
        will be modified in-place.

        .. warning::
            This is an irreversible in-place operation.

        .. versionadded:: 5.0.3

        Parameters
        ----------
        func : callable
            Function that takes in an array (diffraction pattern) and returns an
            array of the exact same shape, with the same data-type.
        callback : callable or None, optional
            Callable that takes an int between 0 and 99. This can be used for progress update.
        processes : int or None, optional
            Number of parallel processes to use. If ``None``, all available processes will be used.
            In case Single Writer Multiple Reader mode is not available, ``processes`` is ignored.

            .. versionadded:: 5.0.6

        Raises
        ------
        TypeError
            if `func` is not a proper callable
        PermissionError
            if the dataset has not been opened with write access.
        """
        if not callable(func):
            raise TypeError(f"Expected a callable argument, but received {type(func)}")

        if callback is None:
            callback = lambda _: None

        ntimes = len(self.time_points)
        dset = self.diffraction_group["intensity"]

        # We implement parallel diff apply in a separate method
        # because single-threaded diff apply can be written with a
        # placeholder array
        if SWMR_AVAILABLE and (processes != 1):
            transformed = ns.pmap(
                _apply_diff,
                self.time_points,
                processes=processes,
                ntotal=ntimes,
                kwargs=dict(fname=self.filename, func=func),
            )

            # We need to switch SWMR mode ON
            # Note that it cannot be turned OFF
            self.swmr_mode = True

            for index, im in enumerate(transformed):
                dset.write_direct(im, dest_sel=np.s_[:, :, index])
                dset.flush()
                callback(int(100 * index / ntimes))
        else:
            # Create a placeholder numpy array where to load and store the results
            placeholder = np.empty(shape=self.resolution, dtype=dset.dtype, order="C")

            for index, _ in enumerate(self.time_points):
                # NOTE: Using dset.read_direct was causing problems because
                #       the destination had shape (N,N), but read_direct wanted a
                #       destination of shape (N,N,1). This is a new behavior since h5py 3.*
                placeholder[:] = dset[:, :, index]
                placeholder[:] = func(placeholder)
                dset.write_direct(placeholder, dest_sel=np.s_[:, :, index])
                callback(int(100 * index / ntimes))

    @write_access_needed
    @update_center
    @update_equilibrium_pattern
    def mask_apply(self, func):
        """
        Modify the diffraction pattern mask ``m`` to ``func(m)``

        Parameters
        ----------
        func : callable
            Function that takes in the diffraction pattern mask, which evaluates to
            ``True`` on valid pixels, and returns an array of the exact same shape,
            with the same data-type.

        Raises
        ------
        TypeError
            if `func` is not a proper callable, or if the result of ``func(m)`` is not boolean.
        ValueError
            if the result of ``func(m)`` does not have the right shape.
        PermissionError
            if the dataset has not been opened with write access.
        """
        if not callable(func):
            raise TypeError(f"Expected a callable argument, but received {type(func)}")

        old_mask = func(self.valid_mask)

        r = func(old_mask)
        if r.dtype != bool:
            raise TypeError(f"Diffraction pattern masks must be boolean, not {r.dtype}")
        if r.shape != old_mask.shape:
            raise ValueError(
                f"Expected diffraction pattern mask with shape {old_mask.shape}, but got {r.shape}"
            )
        self.experimental_parameters_group["valid_mask"][:] = func(self.valid_mask)

    @write_access_needed
    @update_equilibrium_pattern
    def symmetrize(
        self, mod, center=None, kernel_size=None, callback=None, processes=1
    ):
        """
        Symmetrize diffraction images based on n-fold rotational symmetry.

        .. warning::
            This is an irreversible in-place operation.

        Parameters
        ----------
        mod : int
            Fold symmetry number.
        center : array-like, shape (2,) or None
            Coordinates of the center (in pixels). If None (default), the center will be automatically
            determined.
        kernel_size : float or None, optional
            If not None, every diffraction pattern will be smoothed with a gaussian kernel.
            `kernel_size` is the standard deviation of the gaussian kernel in units of pixels.
        callback : callable or None, optional
            Callable that takes an int between 0 and 99. This can be used for progress update.
        processes : int or None, optional
            Number of parallel processes to use. If ``None``, all available processes will be used.
            In case Single Writer Multiple Reader mode is not available, ``processes`` is ignored.

            .. versionadded:: 5.0.6

        Raises
        ------
        ValueError
            if ``mod`` is not a divisor of 360.
        PermissionError
            if the dataset has not been opened with write access.

        See Also
        --------
        diff_apply : apply an operation to each diffraction pattern one-by-one
        """
        if center is None:
            center = self.center
        # Due to possibility of parallel operation,
        # we can't use lambdas or local functions
        # Therefore, we define _symmetrize below and use it here
        apply = partial(
            _symmetrize,
            mod=mod,
            center=center,
            mask=self.valid_mask,
            kernel_size=kernel_size,
        )
        self.diff_apply(apply, callback=callback, processes=processes)

    @property
    def metadata(self):
        """ Dictionary of the dataset's metadata. Dictionary is sorted alphabetically by keys."""
        meta = {k: getattr(self, k) for k in self.valid_metadata}
        meta["filename"] = self.filename
        meta["time_points"] = tuple(self.time_points)
        meta.update(self.compression_params)

        # Ordered dictionary by keys is easiest to inspect
        return OrderedDict(sorted(meta.items(), key=lambda t: t[0]))

    @property
    def valid_mask(self):
        """ Array that evaluates to True on valid pixels (i.e. not on beam-block, not hot pixels, etc.) """
        return np.array(self.experimental_parameters_group["valid_mask"])

    @property
    def invalid_mask(self):
        """ Array that evaluates to True on invalid pixels (i.e. on beam-block, hot pixels, etc.) """
        return np.logical_not(self.valid_mask)

    @property
    def time_points(self):
        # Time-points are not treated as metadata because
        return np.array(self.experimental_parameters_group["time_points"])

    @property
    def resolution(self):
        """ Resolution of diffraction patterns (px, px) """
        intensity_shape = self.diffraction_group["intensity"].shape
        return tuple(intensity_shape[0:2])

    @write_access_needed
    @update_equilibrium_pattern
    def shift_time_zero(self, shift):
        """
        Insert a shift in time points. Reset the shift by setting it to zero. Shifts are
        not consecutive, so that calling `shift_time_zero(20)` twice will not result
        in a shift of 40ps.

        Parameters
        ----------
        shift : float
            Shift [ps]. A positive value of `shift` will move all time-points forward in time,
            whereas a negative value of `shift` will move all time-points backwards in time.

        Raises
        ------
        PermissionError
            if the dataset has not been opened with write access.
        """
        differential = shift - self.time_zero_shift
        self.time_zero_shift = shift
        self.experimental_parameters_group["time_points"][:] = (
            self.time_points + differential
        )

    def _get_time_index(self, timedelay):
        """
        Returns the index of the closest available time-point.

        Parameters
        ----------
        timdelay : float
            Time-delay [ps]

        Returns
        -------
        tp : index
            Index of the Time-point closest to `timedelay` [ps]
        """
        # time_index cannot be cast to int() if np.argwhere returns an empty array
        # catch the corresponding TypeError
        try:
            time_index = int(np.argwhere(self.time_points == float(timedelay)))
        except TypeError:
            time_index = np.argmin(np.abs(self.time_points - float(timedelay)))
            warn(
                f"Time-delay {timedelay}ps not available. Using closest-timedelay {self.time_points[time_index]}ps instead"
            )
        return time_index

    def diff_eq(self):
        """
        Returns the averaged diffraction pattern for all times before photoexcitation.
        In case no data is available before photoexcitation, an array of zeros is returned.

        If the dataset was opened with writing access, the result of this function is
        cached to file. It will be recomputed as needed.

        Time-zero can be adjusted using the ``shift_time_zero`` method.

        Returns
        -------
        I : ndarray, ndim 2
            Diffracted intensity [counts]
        """
        try:
            # The reason this diffraction group might not exist is because
            # this dataset was not part of the initial iris v5 format.
            # it was added in 5.3.0. Therefore, we need to be prepared
            # in case it does not exist in older DiffractionDatasets
            return np.array(self.diffraction_group["equilibrium"])

        # Soon this will become a real error
        except KeyError:
            # Only with write access can diff_eq be cached.
            if self.mode == "r+":
                self._recompute_diff_eq()
                return np.array(self.diffraction_group["equilibrium"])

            # Otherwise, it needs to be calculated from scratch
            intensity = self.diffraction_group["intensity"]
            t0_index = np.argmin(np.abs(self.time_points))

            # If there are no available data before time-zero, np.mean()
            # will return an array of NaNs; instead, return zeros.
            if t0_index == 0:
                return np.zeros(shape=self.resolution, dtype=intensity.dtype)

            return ns.average((intensity[:, :, i] for i in range(t0_index)), axis=2)

    @write_access_needed
    def _recompute_diff_eq(self):
        """ Calculate and store the equilibrium diffraction pattern. """

        intensity = self.diffraction_group["intensity"]
        t0_index = np.argmin(np.abs(self.time_points))

        # If there are no available data before time-zero, np.mean()
        # will return an array of NaNs; instead, return zeros.
        if t0_index == 0:
            diff_eq = np.zeros(shape=self.resolution, dtype=float)
        else:
            diff_eq = ns.average((intensity[:, :, i] for i in range(t0_index)), axis=2)

        eq_dset = self.diffraction_group.require_dataset(
            name="equilibrium", shape=diff_eq.shape, dtype=float
        )
        eq_dset[:] = diff_eq

    def diff_data(self, timedelay, relative=False, out=None):
        """
        Returns diffraction data at a specific time-delay.

        Parameters
        ----------
        timdelay : float or None
            Timedelay [ps]. If None, the entire block is returned.
        relative : bool, optional
            If True, data is returned relative to the average of all diffraction patterns
            before photoexcitation.
        out : ndarray or None, optional
            If an out ndarray is provided, h5py can avoid
            making intermediate copies.

        Returns
        -------
        arr : ndarray
            Time-delay data. If ``out`` is provided, ``arr`` is a view
            into ``out``.

        Raises
        ------
        ValueError
            If timedelay does not exist.
        """
        dataset = self.diffraction_group["intensity"]

        if timedelay is None:
            if out is None:
                out = np.empty_like(dataset)
            dataset.read_direct(out)

        else:
            time_index = self._get_time_index(timedelay)
            if out is None:
                out = np.empty(self.resolution, dtype=dataset.dtype)
            # NOTE: Using dataset.read_direct was causing problems because
            #       the destination had shape (N,N), but read_direct wanted a
            #       destination of shape (N,N,1). This is a new behavior since h5py 3.*
            out[:] = dataset[:, :, time_index]

        if relative:
            out -= self.diff_eq()
            out /= self.diff_eq()

            # Division might introduce infs and nans
            out[:] = np.nan_to_num(out, copy=False)
            np.minimum(out, 2 ** 16 - 1, out=out)

        return out

    def time_series(self, rect, relative=False, out=None):
        """
        Integrated intensity over time inside bounds.

        Parameters
        ----------
        rect : 4-tuple of ints
            Bounds of the region in px. Bounds are specified as [row1, row2, col1, col2]
        relative : bool, optional
            If True, data is returned relative to the average of all diffraction patterns
            before photoexcitation.
        out : ndarray or None, optional
            1-D ndarray in which to store the results. The shape
            should be compatible with ``(len(time_points),)``

        Returns
        -------
        out : ndarray, ndim 1

        See also
        --------
        time_series_selection : intensity integration using arbitrary selections.
        """
        x1, x2, y1, y2 = rect
        data = self.diffraction_group["intensity"][x1:x2, y1:y2, :]
        if relative:
            data -= self.diff_eq()[x1:x2, y1:y2, None]
        return np.mean(data, axis=(0, 1), out=out)

    def time_series_selection(self, selection, relative=False, out=None):
        """
        Integrated intensity over time according to some arbitrary selection. This
        is a generalization of the ``DiffractionDataset.time_series`` method, which
        is much faster, but limited to rectangular selections.

        .. versionadded:: 5.2.1

        Parameters
        ----------
        selection : skued.Selection or ndarray, dtype bool, shape (N,M)
            A selection mask that dictates the regions to integrate in each scattering patterns.
            In the case `selection` is an array, an ArbirarySelection will be used. Performance
            may be degraded. Selection mask evaluating to ``True`` in the regions to integrate.
            The selection must be the same shape as one scattering pattern (i.e. two-dimensional).
        relative : bool, optional
            If True, data is returned relative to the average of all diffraction patterns
            before photoexcitation.
        out : ndarray or None, optional
            1-D ndarray in which to store the results. The shape
            should be compatible with ``(len(time_points),)``

        Returns
        -------
        out : ndarray, ndim 1

        Raises
        ------
        ValueError
            if the shape of ``mask`` does not match the scattering patterns.

        See also
        --------
        time_series : integrated intensity in a rectangle.
        """
        if not isinstance(selection, Selection):
            selection = ArbitrarySelection(selection)

        if selection.shape != self.resolution:
            raise ValueError(
                f"selection mask shape {selection.shape} does not match scattering pattern shape {self.resolution}"
            )

        if out is None:
            out = np.zeros(shape=(len(self.time_points),), dtype=float)

        # For performance reasons, we want to know what is the largest bounding box that
        # fits this selection. Otherwise, all data must be loaded from disk, all the time.
        r1, r2, c1, c2 = selection.bounding_box
        reduced_selection = np.asarray(selection)[r1:r2, c1:c2]

        # There is no way to select data from HDF5 using arbitrary boolean mask
        # Therefore, we must iterate through all time-points.
        dataset = self.diffraction_group["intensity"]
        placeholder = np.empty(shape=(r2 - r1, c2 - c1), dtype=dataset.dtype)
        for index, _ in enumerate(self.time_points):
            # NOTE: Using dataset.read_direct was causing problems because
            #       the destination had shape (N,N), but read_direct wanted a
            #       destination of shape (N,N,1). This is a new behavior since h5py 3.*
            placeholder[:] = dataset[r1:r2, c1:c2, index]

            out[index] = np.mean(placeholder[reduced_selection])

        if relative:
            out -= np.mean(self.diff_eq()[selection])

        return out

    @write_access_needed
    def _autocenter(self):
        """
        Determine the diffraction pattern center automatically.

        .. versionadded:: 5.3.0

        Raises
        ------
        PermissionError
            if the dataset has not been opened with write access.
        ValueError
            If all pixels that are deemed valid have zero intensity.
        """
        intensity = self.diffraction_group["intensity"]
        image = ns.average(intensity[:, :, i] for i in range(intensity.shape[2]))
        if np.allclose(image * self.valid_mask, 0):
            raise ValueError(
                "There is not enough data to determine a center; all valid pixels have zero intensity."
            )
        r, c = autocenter(im=image, mask=self.valid_mask)

        # Note that for backwards-compatibility, the center
        # coordinates need to be stored as (col, row)
        self.center = (c, r)

    @property
    def experimental_parameters_group(self):
        return self.require_group(name=self._exp_params_group_name)

    @property
    def diffraction_group(self):
        return self.require_group(name=self._diffraction_group_name)

    @property
    def compression_params(self):
        """ Compression options in the form of a dictionary """
        dataset = self.diffraction_group["intensity"]
        ckwargs = dict()
        ckwargs["compression"] = dataset.compression
        ckwargs["fletcher32"] = dataset.fletcher32
        ckwargs["shuffle"] = dataset.shuffle
        ckwargs["chunks"] = True if dataset.chunks else None
        if dataset.compression_opts:  # could be None
            ckwargs.update(dataset.compression_opts)
        return ckwargs


# Functions to be passed to pmap must not be local functions
def _apply_diff(timedelay, fname, func):
    with DiffractionDataset(fname, mode="r", libver="latest", swmr=True) as dset:
        im = dset.diff_data(timedelay)
    return func(im)


def _symmetrize(im, mod, center, mask, kernel_size):
    im = nfold(im, mod=mod, center=center, mask=mask)
    if kernel_size is None:
        return im
    return gaussian_filter(im, order=0, sigma=kernel_size, mode="nearest")



# -*- coding: utf-8 -*-
"""
Dialog for symmetrization of DiffractionDataset
"""
from os import cpu_count
from skued import autocenter
import numpy as np
import pyqtgraph as pg
from PyQt5 import QtCore, QtGui, QtWidgets
from .controller import WorkThread
from .qbusyindicator import QBusyIndicator

description = (
    """Align the circle so that its center is aligned with the diffraction center. """
)


class CircleROIWithCenter(pg.CircleROI):

    # Calculating the center position assumes that PyQtGraph is configured
    # such that imageAxisOrder == 'row-major'
    def center():
        corner_x, corner_y = self.pos().x(), self.pos().y()
        radius = self.size().x() / 2
        return (round(corner_x + radius), round(corner_y + radius))

    def set_center(self, x, y):
        radius = self.size().x() / 2
        left = x - radius
        bottom = y - radius
        self.setPos(left, bottom)


class SymmetrizeDialog(QtWidgets.QDialog):
    """
    Modal dialog used to symmetrize datasets.
    """

    error_message_signal = QtCore.pyqtSignal(str)
    symmetrize_parameters_signal = QtCore.pyqtSignal(str, dict)

    def __init__(self, image, mask, center=None, **kwargs):
        super().__init__(**kwargs)
        self.setModal(True)
        self.setWindowTitle("Symmetrization")

        # For use with autocenter
        self._image = image
        self._mask = mask

        title = QtWidgets.QLabel("<h2>Symmetrization Options<\\h2>")
        title.setTextFormat(QtCore.Qt.RichText)
        title.setAlignment(QtCore.Qt.AlignCenter)

        description_label = QtWidgets.QLabel(description, parent=self)
        description_label.setWordWrap(True)
        description_label.setAlignment(QtCore.Qt.AlignCenter)

        self.viewer = pg.ImageView(parent=self)
        self.viewer.setSizePolicy(
            QtWidgets.QSizePolicy.MinimumExpanding,
            QtWidgets.QSizePolicy.MinimumExpanding,
        )
        self.viewer.setImage(image)
        self.center_finder = CircleROIWithCenter(
            pos=np.array(image.shape) / 2 - 100, size=[200, 200], pen=pg.mkPen("r")
        )
        if center is not None and center != (0, 0):
            self.center_finder.set_center(*center)

        self.viewer.getView().addItem(self.center_finder)

        self.mod_widget = QtWidgets.QComboBox(parent=self)
        self.mod_widget.addItems(["2", "3", "4", "6"])
        self.mod_widget.setSizePolicy(
            QtWidgets.QSizePolicy.Maximum, QtWidgets.QSizePolicy.Maximum
        )

        self.smoothing_kernel_widget = QtWidgets.QSpinBox(parent=self)
        self.smoothing_kernel_widget.setRange(0, 100)
        self.smoothing_kernel_widget.setValue(5)
        self.smoothing_kernel_widget.setSuffix(" px")
        self.smoothing_kernel_widget.setEnabled(False)

        self.enable_smoothing_widget = QtWidgets.QCheckBox("Enable gaussian smoothing")
        self.enable_smoothing_widget.setChecked(False)
        self.enable_smoothing_widget.toggled.connect(
            self.smoothing_kernel_widget.setEnabled
        )

        self.processes_widget = QtWidgets.QSpinBox(parent=self)
        self.processes_widget.setRange(1, cpu_count() - 1)
        self.processes_widget.setValue(1)

        self.autocenter_btn = QtWidgets.QPushButton("Autocenter", self)
        self.autocenter_btn.clicked.connect(self.initiate_autocenter)
        self.autocenter_btn.setSizePolicy(
            QtWidgets.QSizePolicy.Maximum, QtWidgets.QSizePolicy.Maximum
        )
        self.busy_indicator = QBusyIndicator(parent=self)
        autocenter_layout = QtWidgets.QHBoxLayout()
        autocenter_layout.addWidget(self.autocenter_btn)
        autocenter_layout.addWidget(self.busy_indicator)

        self.accept_btn = QtWidgets.QPushButton("Symmetrize", self)
        self.accept_btn.clicked.connect(self.accept)
        self.accept_btn.setSizePolicy(
            QtWidgets.QSizePolicy.Maximum, QtWidgets.QSizePolicy.Maximum
        )

        self.cancel_btn = QtWidgets.QPushButton("Cancel", self)
        self.cancel_btn.clicked.connect(self.reject)
        self.cancel_btn.setDefault(True)
        self.cancel_btn.setSizePolicy(
            QtWidgets.QSizePolicy.Maximum, QtWidgets.QSizePolicy.Maximum
        )

        self.error_message_signal.connect(self.show_error_message)

        btns = QtWidgets.QHBoxLayout()
        btns.addWidget(self.accept_btn)
        btns.addWidget(self.cancel_btn)
        btns.addWidget(self.busy_indicator)

        multiplicity_layout = QtWidgets.QFormLayout()
        multiplicity_layout.addRow("Rotational multiplicity: ", self.mod_widget)
        multiplicity_layout.addRow("Number of CPU cores:", self.processes_widget)

        smoothing_layout = QtWidgets.QFormLayout()
        smoothing_layout.addRow(self.enable_smoothing_widget)
        smoothing_layout.addRow(
            "Kernel standard deviation: ", self.smoothing_kernel_widget
        )

        params_layout = QtWidgets.QVBoxLayout()
        params_layout.addWidget(title)
        params_layout.addWidget(description_label)
        params_layout.addLayout(autocenter_layout)
        params_layout.addLayout(multiplicity_layout)
        params_layout.addLayout(smoothing_layout)
        params_layout.addLayout(btns)

        params_widget = QtWidgets.QFrame(parent=self)
        params_widget.setLayout(params_layout)
        params_widget.setFrameShadow(QtWidgets.QFrame.Sunken)
        params_widget.setFrameShape(QtWidgets.QFrame.Panel)
        params_widget.setSizePolicy(
            QtWidgets.QSizePolicy.Maximum, QtWidgets.QSizePolicy.Maximum
        )

        right_layout = QtWidgets.QVBoxLayout()
        right_layout.addWidget(params_widget)
        right_layout.addStretch()

        self.layout = QtWidgets.QHBoxLayout()
        self.layout.addWidget(self.viewer)
        self.layout.addLayout(right_layout)
        self.setLayout(self.layout)

        self.initiate_autocenter()

    @QtCore.pyqtSlot(str)
    def show_error_message(self, msg):
        self.error_dialog = QtGui.QErrorMessage(parent=self)
        self.error_dialog.showMessage(msg)

    @QtCore.pyqtSlot()
    def accept(self):
        self.file_dialog = QtWidgets.QFileDialog(parent=self)
        filename = self.file_dialog.getSaveFileName(filter="*.hdf5")[0]
        if filename == "":
            return

        center = self.center_finder.center()

        params = {
            "center": center,
            "mod": int(self.mod_widget.currentText()),
            "processes": self.processes_widget.value(),
        }

        if self.enable_smoothing_widget.isChecked():
            params["kernel_size"] = self.smoothing_kernel_widget.value()

        self.symmetrize_parameters_signal.emit(filename, params)
        super().accept()

    @QtCore.pyqtSlot()
    def initiate_autocenter(self):
        """Automatically determine the center of an image
        and move the center-finder accordingly"""
        self._worker = AutocenteringThread(
            function=autocenter, kwargs=dict(im=self._image, mask=self._mask)
        )
        self._worker.results_signal.connect(self.set_center)
        self._worker.in_progress_signal.connect(self.busy_indicator.toggle_animation)
        self._worker.in_progress_signal.connect(self.autocenter_btn.setDisabled)
        self._worker.start()

    @QtCore.pyqtSlot(object)
    def set_center(self, rc):
        row, col = rc
        self.center_finder.set_center(col, row)


class AutocenteringThread(WorkThread):
    results_signal = QtCore.pyqtSignal(object)



# -*- coding: utf-8 -*-

import argparse
import sys
import webbrowser
from pathlib import Path
from multiprocessing import freeze_support

from iris import __version__
from iris.gui import run

DESCRIPTION = """Iris is both a library for interacting with ultrafast electron 
diffraction data, as well as a GUI frontend for interactively exploring this data.

Below are some helpful commands. """

EPILOG = """Running this command without any parameters will 
launch the graphical user interface."""

OPEN_HELP = """Dataset to open with iris start-up. """

DOCS_HELP = """Open online documentation in your default web browser."""

parser = argparse.ArgumentParser(prog="iris", description=DESCRIPTION, epilog=EPILOG)
parser.add_argument("-v", "--version", action="version", version=__version__)

subparsers = parser.add_subparsers(
    title="Subcommands", help="Available sub-commands", dest="subcmd"
)

# Parser to open a path
# To facilitate format determination, we need flags specifying whether the
# path points to a raw, compact, or reduced dataset
open_parser = subparsers.add_parser("open", help=OPEN_HELP)
open_parser.add_argument(
    "path", help="Path to the dataset", type=Path, nargs="?", default=None
)
dset_modes = open_parser.add_mutually_exclusive_group(required=True)
dset_modes.add_argument(
    "--raw",
    action="store_true",
    help="This flag indicates that the path should be considered a raw dataset. Raw dataset format will be inferred from the installed plugins",
)
dset_modes.add_argument(
    "--reduced",
    action="store_true",
    help="This flag indicates that the path should be considered a reduced dataset.",
)

# Parser to reach documentation
docs_parser = subparsers.add_parser("docs", help=DOCS_HELP)


def main():
    # This is to support the pynsist-built executables
    # as described here:
    #   https://docs.python.org/3/library/multiprocessing.html#multiprocessing.freeze_support
    freeze_support()

    args = parser.parse_args()

    if args.subcmd == "open":

        # Otherwise, default behavior
        sys.exit(run(path=args.path))

    elif args.subcmd == "docs":
        webbrowser.open("https://iris-ued.readthedocs.io")
        sys.exit(0)

    # Default behavior : open gui without loading any data
    else:
        sys.exit(run(path=None))


if __name__ == "__main__":
    main()



# -*- coding: utf-8 -*-
"""
Broken plug-in
==============
"""

from iris import AbstractRawDataset
import _______ as this_is_an_error


class BrokenPlugin(AbstractRawDataset):
    def __init__(self, source=None, metadata=dict()):
        pass



# -*- coding: utf-8 -*-
from pathlib import Path
from tempfile import gettempdir
from contextlib import redirect_stdout
import io

from iris import load_plugin, AbstractRawDataset, DiffractionDataset

TEST_PLUGIN_PATH = Path(__file__).parent / "plugin_fixture.py"
BROKEN_PLUGIN_PATH = Path(__file__).parent / "broken_plugin.py"


def test_plugin_experimental_parameters():
    """ Test that arbitrary experimental parameters can be manipulated """
    load_plugin(TEST_PLUGIN_PATH)
    from iris.plugins import TestRawDatasetPlugin

    test = TestRawDatasetPlugin()

    test.is_useful = False
    assert not test.is_useful

    test.is_useful = True
    assert test.is_useful


def test_plugin_reduction():
    """ Test that data reduction works """
    load_plugin(TEST_PLUGIN_PATH)
    from iris.plugins import TestRawDatasetPlugin

    test = TestRawDatasetPlugin()
    test.scans = [1, 2]
    test.time_points = [-1, 0, 1]

    temp_file = Path(gettempdir()) / "plugin_test.hdf5"
    with DiffractionDataset.from_raw(test, filename=temp_file, mode="w") as dataset:
        assert test.temperature == dataset.temperature
        # Assert that extra metadata is not kept
        assert hasattr(test, "is_useful")
        assert not hasattr(dataset, "is_useful")


def test_loading_broken_plugin():
    """ Test that exceptions are caught when loading a broken plug-in. """
    f = io.StringIO()
    with redirect_stdout(f):
        load_plugin(BROKEN_PLUGIN_PATH)



# -*- coding: utf-8 -*-
from iris import AbstractRawDataset, check_raw_bounds
from iris.meta import ExperimentalParameter
import numpy as np


class TestRawDataset(AbstractRawDataset):
    """ Class for using raw datasets in tests """

    # We don't want pytest to collect this class as a test
    # https://stackoverflow.com/a/63430765
    __test__ = False

    test = ExperimentalParameter("test", int, default=0)
    resolution = ExperimentalParameter("resolution", tuple, (16, 16))

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.time_points = list(range(0, 10))
        self.scans = list(range(1, 3))

    @check_raw_bounds
    def raw_data(self, timedelay, scan=1):
        return np.ones((self.resolution), dtype=np.uint8) + np.random.randint(
            0, 100, size=self.resolution, dtype=np.uint8
        )
